# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

.. module:: test_vpn_instance
   :synopsis: module that defines several test cases for the vpn_instance
              module.
   TestInitVPNInstance defines a class dedicated to test init a VPNInstance
   with optional vni.
   TestVPNInstance class is dedicated to unit tests for VPNInstance class.
   Setup : Start VPNInstance thread instance.
   TearDown : Stop VPNInstance thread instance.
   VPNInstance is a base class for objects used to manage an E-VPN instance
   (EVI) or IP-VPN instance (VRF)
   Tests are organized as follow :
   - testAx use cases to test endpoints plug with different combinations of MAC
     and IP addresses on a port
   - testBx use cases to test endpoints plug with different combinations of MAC
     and IP addresses on different ports
   - testCx use cases to test endpoints unplug with different combinations of
     MAC and IP addresses as the one plugged on a port
   - testDx use cases to test endpoints unplug with different combinations of
     MAC and IP addresses as the ones plugged on different ports

"""

import mock
import testtools

from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import flowspec
from networking_bagpipe.bagpipe_bgp.engine import ipvpn as ipvpn_routes
from networking_bagpipe.bagpipe_bgp.engine import worker
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers
from networking_bagpipe.bagpipe_bgp.vpn import identifier_allocators
from networking_bagpipe.bagpipe_bgp.vpn import ipvpn
from networking_bagpipe.bagpipe_bgp.vpn import vpn_instance
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t


MAC1 = "00:00:de:ad:be:ef"
IP1 = "10.0.0.1/32"
LOCAL_PORT1 = {'linuxif': 'tap1'}

MAC2 = "00:00:fe:ed:fa:ce"
IP2 = "10.0.0.2/32"
LOCAL_PORT2 = {'linuxif': 'tap2'}

MAC3 = "00:00:de:ad:c0:de"
IP3 = "10.0.0.3/32"
LOCAL_PORT3 = {'linuxif': 'tap3'}

MAC4 = "00:00:fe:ed:f0:0d"
IP4 = "10.0.0.4/32"
LOCAL_PORT4 = {'linuxif': 'tap4'}

RTRecord1 = exa.RTRecord.from_rt(t.RT1)
RTRecord2 = exa.RTRecord.from_rt(t.RT2)
RTRecord3 = exa.RTRecord.from_rt(t.RT3)
RTRecord4 = exa.RTRecord.from_rt(t.RT4)

VPN_ID = 1
VPN_EXT_ID = 1
GW_IP = "10.0.0.1"
GW_MASK = 24
VNID = 255


def _extract_nlri_from_call(vpn_inst, method, call_index=0):
    calls = getattr(vpn_inst, method).call_args_list
    return calls[call_index][0][0].nlri


def _extract_rt_from_call(vpn_inst, method, call_index=0):
    calls = getattr(vpn_inst, method).call_args_list
    return calls[call_index][0][0].route_targets


def _extract_rtrec_from_call(vpn_inst, method, call_index=0):
    calls = getattr(vpn_inst, method).call_args_list
    route = calls[call_index][0][0]
    return route.ecoms(exa.RTRecord)


def _extract_traffic_redirect_from_call(vpn_inst, method, call_index=0):
    calls = getattr(vpn_inst, method).call_args_list
    route = calls[call_index][0][0]
    for ecom in route.ecoms(exa.TrafficRedirect):
        return exa.RouteTarget(int(ecom.asn), int(ecom.target))
    return None


def _extract_traffic_classifier_from_call(vpn_inst, method, call_index=0):
    calls = getattr(vpn_inst, method).call_args_list
    traffic_classifier = vpn_instance.TrafficClassifier()
    traffic_classifier.map_redirect_rules_2_traffic_classifier(
        calls[call_index][0][0].nlri.rules)
    return traffic_classifier


class TestableVPNInstance(vpn_instance.VPNInstance):

    afi = exa.AFI(exa.AFI.ipv4)
    safi = exa.SAFI(exa.SAFI.mpls_vpn)

    def best_route_removed(self, entry, route):
        pass

    def new_best_route(self, entry, route, last):
        pass

    def route_to_tracked_entry(self, route):
        return route

    def generate_vif_bgp_route(self):
        pass


API_PARAMS = {
    'vpn_type': 'EVPN',
    'vpn_instance_id': 'testinstance',
    'mac_address': 'de:ad:00:00:be:ef',
    'ip_address': '192.168.0.1/24',
    'import_rt': ['64512:47'],
    'export_rt': ['64512:47'],
    'local_port': 'tap42'
}


def api_params():
    # return a new dict each time
    # to avoid concurrency issues
    return dict(API_PARAMS)


class TestVPNInstanceAPIChecks(testtools.TestCase):

    def _test_validate_convert_missing(self, method, missing_param,
                                       params=None):
        if params is None:
            params = api_params()
        params.pop(missing_param)
        self.assertRaises(exc.APIMissingParameterException, method, params)

    def test_validate_convert_attach(self):
        method = vpn_instance.VPNInstance.validate_convert_attach_params
        self._test_validate_convert_missing(method, 'vpn_instance_id')
        self._test_validate_convert_missing(method, 'mac_address')
        self._test_validate_convert_missing(method, 'local_port')
        self._test_validate_convert_missing(method, 'import_rt')
        self._test_validate_convert_missing(method, 'export_rt')

    def test_validate_convert_detach(self):
        method = vpn_instance.VPNInstance.validate_convert_detach_params
        self._test_validate_convert_missing(method, 'vpn_instance_id')
        self._test_validate_convert_missing(method, 'mac_address')
        self._test_validate_convert_missing(method, 'local_port')

    def test_api_internal_translation(self):
        params = api_params()
        vpn_instance.VPNInstance.validate_convert_attach_params(params)
        self.assertIn('external_instance_id', params)
        self.assertIn('import_rts', params)
        self.assertIn('export_rts', params)
        self.assertIn('localport', params)

    def test_check_vrf_gateway_ip(self):
        params = api_params()
        params['vpn_type'] = 'IPVPN'
        params['gateway_ip'] = '1.1.1.1'
        ipvpn.VRF.validate_convert_attach_params(params)
        self._test_validate_convert_missing(
            ipvpn.VRF.validate_convert_attach_params,
            'gateway_ip',
            params)

    def test_direction(self):
        params = api_params()
        vpn_instance.VPNInstance.validate_convert_attach_params(params)

    def test_direction_none(self):
        params = api_params()
        params['direction'] = None
        vpn_instance.VPNInstance.validate_convert_attach_params(params)

    def test_direction_ok(self):
        params = api_params()
        params['direction'] = 'to-port'
        vpn_instance.VPNInstance.validate_convert_attach_params(params)

        params = api_params()
        params['direction'] = 'from-port'
        vpn_instance.VPNInstance.validate_convert_attach_params(params)

    def test_direction_bogus(self):
        params = api_params()
        params['direction'] = 'floop'
        self.assertRaises(
            exc.APIException,
            vpn_instance.VPNInstance.validate_convert_attach_params,
            params)

    def test_mac_address_bogus(self):
        params = api_params()
        params['mac_address'] = 'gg:gg:gg:gg:gg:gg'
        self.assertRaises(
            exc.MalformedMACAddress,
            vpn_instance.VPNInstance.validate_convert_attach_params,
            params)

    def test_ip_address_bogus(self):
        params = api_params()
        params['ip_address'] = '257.303.1.'
        self.assertRaises(
            exc.MalformedIPAddress,
            vpn_instance.VPNInstance.validate_convert_attach_params,
            params)


class TestInitVPNInstance(testtools.TestCase):

    def setUp(self):
        super(TestInitVPNInstance, self).setUp()
        self.mock_manager = mock.Mock()
        self.mock_manager.label_allocator.release = mock.Mock()
        self.mock_dp_driver = mock.Mock()
        self.mock_dp_driver.initialize_dataplane_instance = mock.Mock()

    def test_init_stop_VPNInstance_with_forced_vni(self):
        # Initialize a VPNInstance with a forced VNID > 0
        vpn = TestableVPNInstance(self.mock_manager, self.mock_dp_driver,
                                  VPN_EXT_ID, VPN_ID,
                                  [t.RT1], [t.RT1],
                                  GW_IP, GW_MASK,
                                  None, None, vni=VNID)

        # Check that forced VNID is used as instance_label
        self.assertTrue(vpn.forced_vni)
        self.assertEqual(VNID, vpn.instance_label,
                         "VPN instance label should be forced to VNID")
        vpn.dp_driver.initialize_dataplane_instance.assert_called_once_with(
            VPN_ID, VPN_EXT_ID, GW_IP, GW_MASK, VNID)

        # Stop the VPNInstance to check that label release is not called
        vpn.stop()
        vpn.manager.label_allocator.release.assert_not_called()

    def test_init_stop_VPNInstance_without_forced_vni(self):
        # Initialize a VPNInstance with no vni
        vpn = TestableVPNInstance(self.mock_manager, self.mock_dp_driver,
                                  VPN_EXT_ID, VPN_ID, [t.RT1], [t.RT1],
                                  GW_IP, GW_MASK, None, None)

        # Check that VPN instance_label is locally-assigned
        self.assertFalse(vpn.forced_vni)
        vpn.dp_driver.initialize_dataplane_instance.assert_called_once_with(
            VPN_ID, VPN_EXT_ID, GW_IP, GW_MASK, vpn.instance_label)

        # Stop the VPNInstance to check that label release is called
        # with locally assigned instance label
        vpn.stop()
        vpn.manager.label_allocator.release.assert_called_once_with(
            vpn.instance_label)


class TestVPNInstance(t.BaseTestBagPipeBGP, testtools.TestCase):

    def setUp(self):
        super(TestVPNInstance, self).setUp()

        self.mock_dataplane = mock.Mock(
            spec=dataplane_drivers.VPNInstanceDataplane)

        mock_dp_driver = mock.Mock(
            spec=dataplane_drivers.DataplaneDriver)
        mock_dp_driver.initialize_dataplane_instance.return_value = (
            self.mock_dataplane
        )

        self.vpn = TestableVPNInstance(mock.Mock(name='VPNManager'),
                                       mock_dp_driver, 1, 1,
                                       [t.RT1], [t.RT1], '10.0.0.1', 24,
                                       None, None)

        self.vpn.synthesize_vif_bgp_route = mock.Mock(
            return_value=engine.RouteEntry(t.NLRI1, [t.RT1]))
        self.vpn._advertise_route = mock.Mock()
        self.vpn._withdraw_route = mock.Mock()
        self.vpn.start()

        self.set_event_target_worker(self.vpn)

    def tearDown(self):
        super(TestVPNInstance, self).tearDown()
        self.vpn.stop()
        self.vpn.join()

    def _get_ip_address(self, ip_address_prefix):
        return ip_address_prefix[0:ip_address_prefix.find('/')]

    def _validate_ip_address_2_mac_address_consistency(self, mac_address,
                                                       ip_address1,
                                                       ip_address2=None):
        # Validate IP address -> MAC address consistency
        self.assertIn(ip_address1, self.vpn.ip_address_2_mac)

        if ip_address2:
            self.assertIn(ip_address1, self.vpn.ip_address_2_mac)
            self.assertEqual(
                self.vpn.ip_address_2_mac[ip_address1],
                self.vpn.ip_address_2_mac[ip_address2])
        else:
            self.assertIn(
                mac_address, self.vpn.ip_address_2_mac[ip_address1])

    def _chk_mac_2_localport_data_consistency(self, mac_address, localport):
        # Validate MAC address -> Port informations consistency
        self.assertIn(mac_address, self.vpn.mac_2_localport_data)

        port_info = self.vpn.mac_2_localport_data[
            mac_address]['port_info']
        self.assertEqual(localport['linuxif'], port_info['linuxif'])

    def _validate_localport_2_endpoints_consistency(self, length, localport,
                                                    endpoints):
        # Validate Port -> Endpoint (MAC, IP) tuple consistency
        self.assertEqual(
            length,
            len(self.vpn.localport_2_endpoints[localport['linuxif']]))

        for endpoint in endpoints:
            self.assertIn(
                endpoint,
                self.vpn.localport_2_endpoints[localport['linuxif']])

    def test_validate_convert_params_duplicate_rts(self):
        test_params = {'vpn_instance_id': 'foo',
                       'mac_address': 'aa:bb:cc:dd:ee:ff',
                       'ip_address': '1.2.3.4',
                       'local_port': 'foo',
                       'import_rt': ['64512:1', '64512:1'],
                       'export_rt': '64512:4, 64512:4'}

        vpn_instance.VPNInstance.validate_convert_params(test_params)
        self.assertEqual(['64512:1'], test_params['import_rt'])
        self.assertEqual(['64512:4'], test_params['export_rt'])

    def test_a1_plug_endpoint_twice_same_port(self):
        # Plug one endpoint with same MAC and IP addresses twice on a port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        self.vpn.dataplane.vif_plugged.assert_called_once()

        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])

    def test_a2_plug_multiple_endpoints_with_same_ip_same_port(self):
        # Plug multiple endpoints with different MAC addresses and same IP
        # address on a port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # An IP address correspond to only one MAC address, exception must be
        # raised
        self.assertRaises(exc.APIException,
                          self.vpn.vif_plugged,
                          MAC2, IP1, LOCAL_PORT1)
        self.vpn.dataplane.vif_plugged.assert_called_once()
        self.vpn._advertise_route.assert_called_once()

        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(MAC2, self.vpn.mac_2_localport_data)

    def test_a3_plug_multiple_endpoints_with_same_mac_same_port(self):
        # Plug multiple endpoints with same MAC address and different IP
        # addresses on a port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC1, IP2, LOCAL_PORT1)

        self.assertEqual(2, self.vpn.dataplane.vif_plugged.call_count,
                         "Port different IP addresses must be plugged on "
                         "dataplane")
        self.assertEqual(2, self.vpn._advertise_route.call_count,
                         "Route for port different IP addresses must be "
                         "advertised")
        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1, IP2)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            2, LOCAL_PORT1, [(MAC1, IP1), (MAC1, IP2)])

    def test_a4_plug_multiple_endpoints_same_port(self):
        # Plug multiple endpoints with different MAC and IP addresses on a port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        self.assertEqual(2, self.vpn.dataplane.vif_plugged.call_count,
                         "Port different endpoints must be plugged on "
                         "dataplane")
        self.assertEqual(2, self.vpn._advertise_route.call_count,
                         "Route for port different endpoints must be "
                         "advertised")
        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._validate_ip_address_2_mac_address_consistency(MAC2, IP2)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._chk_mac_2_localport_data_consistency(MAC2, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            2, LOCAL_PORT1, [(MAC1, IP1), (MAC2, IP2)])

    def test_b1_plug_endpoint_twice_different_port(self):
        # Plug one endpoint with same MAC and IP addresses twice on different
        # ports

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # A port correspond to only one MAC address, exception must be raised
        self.assertRaises(exc.APIException,
                          self.vpn.vif_plugged,
                          MAC1, IP1, LOCAL_PORT2)
        self.vpn.dataplane.vif_plugged.assert_called_once()
        self.vpn._advertise_route.assert_called_once()

        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(
            LOCAL_PORT2['linuxif'], self.vpn.localport_2_endpoints)

    def test_b2_plug_multiple_endpoints_with_same_ip_different_port(self):
        # Plug multiple endpoints with different MAC addresses and same IP
        # address on different port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # An IP address correspond to only one MAC address, exception must be
        # raised
        self.assertRaises(exc.APIException,
                          self.vpn.vif_plugged,
                          MAC2, IP1, LOCAL_PORT2)
        self.vpn.dataplane.vif_plugged.assert_called_once()
        self.vpn._advertise_route.assert_called_once()

        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(
            LOCAL_PORT2['linuxif'], self.vpn.localport_2_endpoints)

    def test_b4_plug_multiple_endpoints_with_same_mac_different_port(self):
        # Plug multiple endpoints with same MAC address and different IP
        # addresses on different ports

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # A port correspond to only one MAC address, exception must be raised
        self.assertRaises(exc.APIException,
                          self.vpn.vif_plugged,
                          MAC1, IP2, LOCAL_PORT2)
        self.vpn.dataplane.vif_plugged.assert_called_once()
        self.vpn._advertise_route.assert_called_once()

        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(
            LOCAL_PORT2['linuxif'], self.vpn.localport_2_endpoints)

    def test_b5_plug_multiple_endpoints_different_port(self):
        # Plug multiple endpoints with different MAC and IP addresses on
        # different ports

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT2)

        self.assertEqual(2, self.vpn.dataplane.vif_plugged.call_count,
                         "All ports must be plugged on dataplane")
        self.assertEqual(2, self.vpn._advertise_route.call_count,
                         "Routes for all ports must be advertised")

        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])

        self._validate_ip_address_2_mac_address_consistency(MAC2, IP2)
        self._chk_mac_2_localport_data_consistency(MAC2, LOCAL_PORT2)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT2, [(MAC2, IP2)])

    def test_c1_unplug_unique_endpoint_same_port(self):
        # Unplug one endpoint with same MAC and IP addresses as the one plugged
        # on port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        label1 = self.vpn.mac_2_localport_data[MAC1]['label']

        self.vpn.vif_unplugged(MAC1, IP1)

        self.vpn.dataplane.vif_unplugged.assert_called_once()
        self.vpn.dataplane.vif_unplugged.assert_called_with(
            MAC1, self._get_ip_address(IP1), LOCAL_PORT1, label1, None, True)
        self.vpn._advertise_route.assert_called_once()
        self.vpn._withdraw_route.assert_called_once()

        self.assertEqual({}, self.vpn.mac_2_localport_data)
        self.assertEqual({}, self.vpn.ip_address_2_mac)
        self.assertEqual({}, self.vpn.localport_2_endpoints)

    def test_c2_unplug_unique_endpoint_with_same_ip_same_port(self):
        # Unplug one endpoint with different MAC addresses and same IP address
        # as the one plugged on port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        self.assertRaises(exc.APIException,
                          self.vpn.vif_unplugged,
                          MAC2, IP1)

        self.vpn.dataplane.vif_unplugged.assert_not_called()
        self.vpn._advertise_route.assert_called_once()

        self.assertIn(MAC1, self.vpn.mac_2_localport_data)
        self.assertIn(IP1, self.vpn.ip_address_2_mac)
        self.assertIn(LOCAL_PORT1['linuxif'], self.vpn.localport_2_endpoints)

    def test_c3_unplug_unique_endpoint_with_same_mac_same_port(self):
        # Unplug one endpoint with same MAC address and different IP addresses
        # as the one plugged on port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        self.assertRaises(exc.APIException,
                          self.vpn.vif_unplugged,
                          MAC1, IP2)

        self.vpn.dataplane.vif_unplugged.assert_not_called()
        self.vpn._advertise_route.assert_called_once()
        self.vpn._withdraw_route.assert_not_called()

        self.assertIn(MAC1, self.vpn.mac_2_localport_data)
        self.assertIn(IP1, self.vpn.ip_address_2_mac)
        self.assertIn(LOCAL_PORT1['linuxif'], self.vpn.localport_2_endpoints)

    def test_c4_unplug_one_endpoint_same_port(self):
        # Unplug only one endpoint with same MAC and IP addresses
        # corresponding to one plugged on port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        label1 = self.vpn.mac_2_localport_data[MAC1]['label']

        self.vpn.vif_unplugged(MAC1, IP1)

        self.vpn.dataplane.vif_unplugged.assert_called_once()
        self.vpn.dataplane.vif_unplugged.assert_called_with(
            MAC1, self._get_ip_address(IP1), LOCAL_PORT1, label1, None, False)
        self.assertEqual(2, self.vpn._advertise_route.call_count,
                         "Routes for all port endpoints must be first "
                         "advertised and only one withdrawn")
        self.vpn._withdraw_route.assert_called_once()

        self._validate_ip_address_2_mac_address_consistency(MAC2, IP2)
        self._chk_mac_2_localport_data_consistency(MAC2, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC2, IP2)])

    def test_c5_unplug_all_endpoints_same_port(self):
        # Unplug all endpoints with same MAC and IP addresses
        # corresponding to those plugged on port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        label1 = self.vpn.mac_2_localport_data[MAC1]['label']
        label2 = self.vpn.mac_2_localport_data[MAC2]['label']

        self.vpn.vif_unplugged(MAC1, IP1)
        self.vpn.vif_unplugged(MAC2, IP2)

        self.assertEqual(2, self.vpn.dataplane.vif_unplugged.call_count,
                         "All port endpoints must be unplugged from dataplane")
        self.vpn.dataplane.vif_unplugged.assert_has_calls([
            mock.call(MAC1, self._get_ip_address(IP1),
                      LOCAL_PORT1, label1, None, False),
            mock.call(MAC2, self._get_ip_address(IP2),
                      LOCAL_PORT1, label2, None, True)
        ])
        self.assertEqual(2, self.vpn._advertise_route.call_count,
                         "Routes for all port endpoints must be first "
                         "advertised and after withdrawn")
        self.assertEqual(2, self.vpn._withdraw_route.call_count,
                         "Routes for all port endpoints must be first "
                         "advertised and after withdrawn")

        self.assertEqual({}, self.vpn.mac_2_localport_data)
        self.assertEqual({}, self.vpn.ip_address_2_mac)
        self.assertEqual({}, self.vpn.localport_2_endpoints)

    def test_d1_unplug_unique_endpoints_different_port(self):
        # Unplug the endpoints with different MAC and IP addresses
        # corresponding to those plugged on different ports

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT2)

        label1 = self.vpn.mac_2_localport_data[MAC1]['label']
        label2 = self.vpn.mac_2_localport_data[MAC2]['label']

        self.vpn.vif_unplugged(MAC1, IP1)
        self.vpn.vif_unplugged(MAC2, IP2)

        self.assertEqual(2, self.vpn.dataplane.vif_unplugged.call_count,
                         "All different ports endpoints must be unplugged "
                         "from dataplane")

        self.vpn.dataplane.vif_unplugged.assert_has_calls([
            mock.call(MAC1, self._get_ip_address(IP1),
                      LOCAL_PORT1, label1, None, True),
            mock.call(MAC2, self._get_ip_address(IP2),
                      LOCAL_PORT2, label2, None, True)
        ])
        self.assertEqual(2, self.vpn._advertise_route.call_count,
                         "Routes for all different ports endpoints must be "
                         "first advertised and after withdrawn")
        self.assertEqual(2, self.vpn._withdraw_route.call_count,
                         "Routes for all different ports endpoints must be "
                         "first advertised and after withdrawn")

        self.assertEqual({}, self.vpn.mac_2_localport_data)
        self.assertEqual({}, self.vpn.ip_address_2_mac)
        self.assertEqual({}, self.vpn.localport_2_endpoints)

    def test_d2_unplug_one_endpoint_same_ip_different_port(self):
        # Unplug one endpoint with different MAC or IP address corresponding to
        # one plugged on another port

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT2)

        self.assertRaises(exc.APIException,
                          self.vpn.vif_unplugged,
                          MAC1, IP2)

        self.vpn.dataplane.vif_unplugged.assert_not_called()
        self.assertEqual(2, self.vpn._advertise_route.call_count,
                         "Routes for all different ports endpoints must only "
                         "be advertised")

        self._validate_ip_address_2_mac_address_consistency(MAC1, IP1)
        self._chk_mac_2_localport_data_consistency(MAC1, LOCAL_PORT1)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])

        self._validate_ip_address_2_mac_address_consistency(MAC2, IP2)
        self._chk_mac_2_localport_data_consistency(MAC2, LOCAL_PORT2)
        self._validate_localport_2_endpoints_consistency(
            1, LOCAL_PORT2, [(MAC2, IP2)])

    def test_d3_unplug_multiple_endpoints_different_port(self):
        # Unplug multiple endpoints with same MAC and IP addresses
        # corresponding to those plugged on different ports

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC3, IP3, LOCAL_PORT2)
        self.vpn.vif_plugged(MAC4, IP4, LOCAL_PORT2)

        label1 = self.vpn.mac_2_localport_data[MAC1]['label']
        label2 = self.vpn.mac_2_localport_data[MAC2]['label']
        label3 = self.vpn.mac_2_localport_data[MAC3]['label']
        label4 = self.vpn.mac_2_localport_data[MAC4]['label']

        self.vpn.vif_unplugged(MAC1, IP1)
        self.vpn.vif_unplugged(MAC2, IP2)
        self.vpn.vif_unplugged(MAC3, IP3)
        self.vpn.vif_unplugged(MAC4, IP4)

        self.assertEqual(4, self.vpn.dataplane.vif_unplugged.call_count,
                         "All different ports endpoints must be unplugged "
                         "from dataplane")
        self.vpn.dataplane.vif_unplugged.assert_has_calls([
            mock.call(MAC1, self._get_ip_address(IP1),
                      LOCAL_PORT1, label1, None, False),
            mock.call(MAC2, self._get_ip_address(IP2),
                      LOCAL_PORT1, label2, None, True),
            mock.call(MAC3, self._get_ip_address(IP3),
                      LOCAL_PORT2, label3, None, False),
            mock.call(MAC4, self._get_ip_address(IP4),
                      LOCAL_PORT2, label4, None, True)
        ])
        self.assertEqual(4, self.vpn._withdraw_route.call_count,
                         "Routes for all different ports endpoints must be "
                         "first advertised and after withdrawn")
        self.assertEqual(4, self.vpn._advertise_route.call_count,
                         "Routes for all different ports endpoints must be "
                         "first advertised and after withdrawn")

        self.assertEqual({}, self.vpn.mac_2_localport_data)
        self.assertEqual({}, self.vpn.ip_address_2_mac)
        self.assertEqual({}, self.vpn.localport_2_endpoints)

    def test_get_lg_localport_data(self):
        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC3, IP3, LOCAL_PORT2)
        self.vpn.vif_plugged(MAC4, IP4, LOCAL_PORT2)

        self.vpn.get_lg_local_port_data("")

    # tests of update_route_targets

    def _test_update_rts_init(self):
        self.vpn._advertise_route.reset_mock()

        route = engine.RouteEntry(t.NLRI1, [t.RT1])
        self.vpn.endpoint_2_route = {None: route}

    def test_update_rts_1(self):
        self._test_update_rts_init()

        # no change -> no route update
        self.vpn.update_route_targets([t.RT1], [t.RT1])

        self.vpn._advertise_route.assert_not_called()

    def test_update_rts_2(self):
        self._test_update_rts_init()

        # change imports -> no route update
        self.vpn.update_route_targets([t.RT2], [t.RT1])

        self.vpn._advertise_route.assert_not_called()

    def test_update_rts_3(self):
        self._test_update_rts_init()

        # change exports
        # check that previously advertised routes are readvertised
        self.vpn.update_route_targets([t.RT1], [t.RT2])

        self.vpn._advertise_route.assert_called_once()

        self.assertIn(t.RT2, _extract_rt_from_call(self.vpn,
                                                   '_advertise_route'))
        self.assertNotIn(t.RT1, _extract_rt_from_call(self.vpn,
                                                      '_advertise_route'))

    def test_update_rts_3bis(self):
        self._test_update_rts_init()

        # change exports
        # check that previously advertised routes are readvertised
        self.vpn.update_route_targets([t.RT1], [t.RT1, t.RT2])

        self.vpn._advertise_route.assert_called_once()
        self.assertIn(t.RT2, _extract_rt_from_call(self.vpn,
                                                   '_advertise_route'))
        self.assertIn(t.RT1, _extract_rt_from_call(self.vpn,
                                                   '_advertise_route'))

    def test_cleanup_assist(self):
        # simulate a route injected in our VPNInstance

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        self._new_route_event(engine.RouteEvent.ADVERTISE,
                              self._fake_nlri("fake NLRI"),
                              [t.RT1, t.RT2], worker_a, t.NH1, 200)

        self.mock_dataplane.needs_cleanup_assist.return_value = True

        with mock.patch.object(self.vpn, 'best_route_removed') as mock_brr:
            self.vpn.stop()
            mock_brr.assert_called_once()

    def test_plug_endpoint_direction_to_port(self):
        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1, direction='to-port')

        self.vpn.dataplane.vif_plugged.assert_called_once()
        self.vpn._advertise_route.assert_called_once()

    def test_plug_endpoint_direction_from_port(self):
        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1, direction='from-port')

        self.vpn.dataplane.vif_plugged.assert_called_once()
        self.vpn._advertise_route.assert_not_called()


LOCAL_ADDRESS = '4.5.6.7'
NEXT_HOP = '45.45.45.45'

IP_ADDR_PREFIX1 = '1.1.1.1/32'
IP_ADDR_PREFIX2 = '2.2.2.2/32'
IP_ADDR_PREFIX3 = '3.3.3.3/32'

STATIC_ADDR_PREFIX1 = '10.10.10.10/32'
STATIC_ADDR_PREFIX2 = '20.20.20.20/32'

ATTRACT_TRAFFIC_1 = {'redirect_rts': [t.RT5],
                     'classifier': {'destinationPort': '80',
                                    'protocol': 'tcp'
                                    }
                     }

ATTRACT_STATIC_1 = {'to_rt': [t.RT4],
                    'static_destination_prefixes': [STATIC_ADDR_PREFIX1],
                    'redirect_rts': [t.RT5],
                    'classifier': {'destinationPort': '80',
                                   'protocol': 'tcp'
                                   }
                    }

ATTRACT_STATIC_2 = {'to_rt': [t.RT4],
                    'static_destination_prefixes': [STATIC_ADDR_PREFIX1,
                                                    STATIC_ADDR_PREFIX2],
                    'redirect_rts': [t.RT5],
                    'classifier': {'destinationPort': '80',
                                   'protocol': 'tcp'
                                   }
                    }

TC1 = vpn_instance.TrafficClassifier(destination_prefix="1.1.1.1/32",
                                     destination_port="80",
                                     protocol="tcp")

TC2 = vpn_instance.TrafficClassifier(destination_prefix="2.2.2.2/32",
                                     destination_port="80",
                                     protocol="tcp")

TC_STATIC1 = vpn_instance.TrafficClassifier(
    destination_prefix=STATIC_ADDR_PREFIX1,
    destination_port="80",
    protocol="tcp"
)

TC_STATIC2 = vpn_instance.TrafficClassifier(
    destination_prefix=STATIC_ADDR_PREFIX2,
    destination_port="80",
    protocol="tcp"
)


class TestVRF(t.BaseTestBagPipeBGP, testtools.TestCase):

    def setUp(self):
        super(TestVRF, self).setUp()

        self.mock_dp = mock.Mock(
            spec=ipvpn.DummyVPNInstanceDataplane)

        mock_dp_driver = mock.Mock(
            spec=ipvpn.DummyDataplaneDriver)

        mock_dp_driver.initialize_dataplane_instance.return_value = \
            self.mock_dp
        mock_dp_driver.get_local_address.return_value = LOCAL_ADDRESS
        mock_dp_driver.supported_encaps.return_value = \
            [exa.Encapsulation(exa.Encapsulation.Type.DEFAULT)]

        label_alloc = identifier_allocators.LabelAllocator()
        bgp_manager = mock.Mock()
        bgp_manager.get_local_address.return_value = LOCAL_ADDRESS
        rd_alloc = (
            identifier_allocators.RDAllocator(bgp_manager.get_local_address())
        )
        self.manager = mock.Mock(bgp_manager=bgp_manager,
                                 label_allocator=label_alloc,
                                 rd_allocator=rd_alloc)

        self.vpn = ipvpn.VRF(self.manager, mock_dp_driver, 1, 1,
                             [t.RT1], [t.RT1], '10.0.0.1', 24,
                             {'from_rt': [t.RT3],
                              'to_rt': [t.RT4]},
                             None)

        self.vpn._advertise_route = mock.Mock()
        self.vpn._withdraw_route = mock.Mock()
        self.vpn.start()

        self.set_event_target_worker(self.vpn)

    def _reset_mocks(self):
        self.vpn._advertise_route.reset_mock()
        self.vpn._withdraw_route.reset_mock()
        self.mock_dp.setup_dataplane_for_remote_endpoint.reset_mock()
        self.mock_dp.vif_plugged.reset_mock()
        self.mock_dp.vif_unplugged.reset_mock()

    def tearDown(self):
        super(TestVRF, self).tearDown()
        self.vpn.stop()
        self.vpn.join()

    def _config_vrf_with_attract_traffic(self, attract_traffic,
                                         no_readvertise=False):
        self.vpn.attract_traffic = True
        self.vpn.attract_rts = attract_traffic['redirect_rts']
        self.vpn.attract_classifier = attract_traffic['classifier']

        if no_readvertise:
            self.vpn.readvertise = False
            self.vpn.readvertise_from_rts = []
            self.vpn.readvertise_to_rts = attract_traffic['to_rt']
        else:
            if attract_traffic.get('to_rt'):
                self.assertEqual(self.vpn.readvertise_to_rts,
                                 attract_traffic['to_rt'])

        if (attract_traffic.get('to_rt') and
                attract_traffic.get('static_destination_prefixes')):
            self.vpn.attract_static_dest_prefixes = (
                attract_traffic['static_destination_prefixes']
            )

    def _mock_vpnmanager_for_attract_traffic(self):
        self.manager.redirect_traffic_to_vpn = mock.Mock(
            spec=ipvpn.VPNInstanceDataplane)
        self.manager.stop_redirect_to_vpn = mock.Mock()

    def _reset_mocks_vpnmanager(self):
        self.manager.redirect_traffic_to_vpn.reset_mock()
        self.manager.stop_redirect_to_vpn.reset_mock()

    def _generate_route_nlri(self, ip_address_prefix, nexthop=NEXT_HOP):
        # Parse address/mask
        (_, prefix_len) = self.vpn._parse_ipaddress_prefix(ip_address_prefix)

        prefix_rd = self.manager.rd_allocator.get_new_rd(
            "Route distinguisher for prefix %s" % ip_address_prefix
        )
        rd = prefix_rd if prefix_len == 32 else self.vpn.instance_rd

        label = self.manager.label_allocator.get_new_label(
            "Label for prefix %s" % ip_address_prefix
        )

        return ipvpn_routes.IPVPNRouteFactory(exa.AFI(exa.AFI.ipv4),
                                              ip_address_prefix,
                                              label, rd, nexthop)

    def _generate_flow_spec_nlri(self, classifier):
        rd = self.manager.rd_allocator.get_new_rd(
            "Route distinguisher for FlowSpec NLRI"
        )
        flow_nlri = flowspec.FlowRouteFactory(exa.AFI(exa.AFI.ipv4), rd)

        for rule in classifier.map_traffic_classifier_2_redirect_rules():
            flow_nlri.add(rule)

        return flow_nlri

    def test_validate_convert_attach(self):
        params = api_params()
        params.pop('ip_address')
        self.assertRaises(exc.APIMissingParameterException,
                          ipvpn.VRF.validate_convert_attach_params,
                          params)

    def test_validate_convert_detach(self):
        params = api_params()
        params.pop('ip_address')
        self.assertRaises(exc.APIMissingParameterException,
                          ipvpn.VRF.validate_convert_detach_params,
                          params)

    # unit test for IPVPN re-advertisement
    def test_re_advertisement_1(self):
        self._reset_mocks()

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri_1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri_1,
                              [t.RT1, t.RT2], worker_a, t.NH1, 200)
        # no re-advertisement supposed to happen
        self.vpn._advertise_route.assert_called_once()
        # dataplane supposed to be updated for this route
        self.mock_dp.setup_dataplane_for_remote_endpoint.assert_called_once()

        self._reset_mocks()

        vpn_nlri_2 = self._generate_route_nlri(IP_ADDR_PREFIX2)
        event2 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri_2,
                                       [t.RT3], worker_a, t.NH1, 200,
                                       rtrecords=[RTRecord1])
        # re-advertisement of VPN NLRI2 supposed to happen, to RT4
        self.vpn._advertise_route.assert_called_once()
        self.assertIn(t.RT4, _extract_rt_from_call(self.vpn,
                                                   '_advertise_route'))
        self.assertNotIn(t.RT2, _extract_rt_from_call(self.vpn,
                                                      '_advertise_route'))
        self.assertNotIn(t.RT3, _extract_rt_from_call(self.vpn,
                                                      '_advertise_route'))
        self.assertIn(RTRecord3, _extract_rtrec_from_call(self.vpn,
                                                          '_advertise_route'))
        self.assertIn(RTRecord1, _extract_rtrec_from_call(self.vpn,
                                                          '_advertise_route'))
        # check that event is for re-advertised route vpn_nlri_2 and
        #  contains what we expect
        route_entry = self.vpn._advertise_route.call_args_list[0][0][0]
        self.assertNotEqual(vpn_nlri_2.rd, route_entry.nlri.rd)
        # dataplane *not* supposed to be updated for this route
        self.mock_dp.setup_dataplane_for_remote_endpoint.assert_not_called()

        self._reset_mocks()

        # new interface plugged in
        # route vpn_nlri_2 should be re-advertized with this new next hop as
        #  next-hop
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT2)
        # advertised route count should increment by 2:
        # - vif route itself
        # - re-adv of NLRI1 with this new port as next-hop
        self.assertEqual(2, self.vpn._advertise_route.call_count)
        self.vpn._withdraw_route.assert_not_called()
        self.assertIn(t.RT1, _extract_rt_from_call(self.vpn,
                                                   '_advertise_route', 0))
        self.assertNotIn(t.RT4, _extract_rt_from_call(self.vpn,
                                                      '_advertise_route', 0))
        self.assertIn(t.RT4, _extract_rt_from_call(self.vpn,
                                                   '_advertise_route', 1))
        self.assertNotIn(t.RT1, _extract_rt_from_call(self.vpn,
                                                      '_advertise_route', 1))

        # check that second event is for re-advertised route vpn_nlri_2 and
        #  contains what we expect
        route_entry = self.vpn._advertise_route.call_args_list[1][0][0]
        vpn_nlri_2_readv_rd = route_entry.nlri.rd
        self.assertEqual(vpn_nlri_2.cidr.prefix(),
                         route_entry.nlri.cidr.prefix())
        self.assertNotEqual(vpn_nlri_2.labels, route_entry.nlri.labels)
        self.assertNotEqual(vpn_nlri_2.nexthop, route_entry.nlri.nexthop)
        self.assertNotEqual(vpn_nlri_2.rd, route_entry.nlri.rd)
        self.assertEqual(vpn_nlri_2_readv_rd, route_entry.nlri.rd)

        self._reset_mocks()

        # new route, that, because it contains the redirectRT in RTRecord
        # will not be re-advertized
        vpn_nlri3 = self._generate_route_nlri(IP_ADDR_PREFIX3)
        event3 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri3,
                                       [t.RT3], worker_a, t.NH1, 200,
                                       rtrecords=[RTRecord4])
        self.vpn._advertise_route.assert_not_called()
        self.vpn._withdraw_route.assert_not_called()
        self._revert_event(event3)

        self._reset_mocks()

        # vif unplugged, routes VPN NLRI2 with next-hop
        # corresponding to this ports should now be withdrawn
        self.vpn.vif_unplugged(MAC2, IP2)
        self.assertEqual(2, self.vpn._withdraw_route.call_count)
        route_entry = self.vpn._withdraw_route.call_args_list[0][0][0]
        self.assertEqual(vpn_nlri_2.cidr.prefix(),
                         route_entry.nlri.cidr.prefix())
        self.assertNotEqual(vpn_nlri_2.labels, route_entry.nlri.labels)
        self.assertNotEqual(vpn_nlri_2.nexthop, route_entry.nlri.nexthop)
        self.assertNotEqual(vpn_nlri_2.rd, route_entry.nlri.rd)
        self.assertNotEqual(vpn_nlri_2.rd, route_entry.nlri.rd)
        self.assertEqual(vpn_nlri_2_readv_rd, route_entry.nlri.rd)

        self._reset_mocks()

        # RTs of route NLRI1 now include a re-advertiseed RT
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri_1,
                              [t.RT1, t.RT2, t.RT3],
                              worker_a, t.NH1, 200)
        self.vpn._advertise_route.assert_called_once()
        self.assertIn(t.RT4, _extract_rt_from_call(self.vpn,
                                                   '_advertise_route'))
        # dataplane supposed to be updated for this route
        self.mock_dp.setup_dataplane_for_remote_endpoint.assert_called_once()

        self._reset_mocks()

        self._revert_event(event2)
        # withdraw of re-adv route supposed to happen
        self.vpn._withdraw_route.assert_called_once()
        self.vpn._advertise_route.assert_not_called()
        # dataplane *not* supposed to be updated for this route
        self.mock_dp.setup_dataplane_for_remote_endpoint.assert_not_called()

    def test_re_advertisement_last(self):
        self._reset_mocks()

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri_1 = self._generate_route_nlri(IP_ADDR_PREFIX1, t.NH1)
        event1 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri_1,
                                       [t.RT3], worker_a, t.NH1, 200,
                                       rtrecords=[RTRecord1])
        # advertised route count increment by 2:
        # - vif route for port
        # - re-advertisement of VPN NLRI1 with port as next-hop
        self.assertEqual(2, self.vpn._advertise_route.call_count)
        self.vpn._withdraw_route.assert_not_called()

        self._reset_mocks()

        vpn_nlri_1bis = self._generate_route_nlri(IP_ADDR_PREFIX1, t.NH1)
        event1bis = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                          vpn_nlri_1bis, [t.RT3], worker_a,
                                          t.NH1, 200, rtrecords=[RTRecord1])
        # second re-advertisement of VPN NLRI1 supposed to happen
        # (must be futher fixed to only append once)
        self.vpn._advertise_route.assert_called_once()
        self.vpn._withdraw_route.assert_not_called()

        self._reset_mocks()

        self._revert_event(event1bis)
        self.vpn._advertise_route.assert_not_called()
        self.vpn._withdraw_route.assert_not_called()

        self._reset_mocks()

        self._revert_event(event1)
        # withdraw of re-adv route supposed to happen
        self.vpn._withdraw_route.assert_called_once()
        self.vpn._advertise_route.assert_not_called()

    # unit test for FlowSpec re-advertisement
    def test_flowspec_re_advertisement_1(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        # FlowSpec route
        flow_nlri1 = self._generate_flow_spec_nlri(TC1)
        self._new_flow_event(engine.RouteEvent.ADVERTISE, flow_nlri1, [t.RT5],
                             [t.RT3], worker_a)

        # re-advertisement of Flow NLRI1 supposed to happen, to RT4
        self.assertEqual(2, self.vpn._advertise_route.call_count)
        self.manager.redirect_traffic_to_vpn.assert_not_called()

        # 1 - re-advertisement of a default route supposed to happen
        # to RT4
        self.assertIn(t.RT4,
                      _extract_rt_from_call(self.vpn, '_advertise_route', 0))

        ipvpn_nlri = _extract_nlri_from_call(self.vpn, '_advertise_route', 0)
        self.assertEqual(ipvpn.DEFAULT_ADDR_PREFIX, ipvpn_nlri.cidr.prefix())

        # 2 - advertisement of FlowSpec NLRI supposed to happen to RT4
        #     for traffic redirection to RT5 on TCP destination port 80
        self.assertIn(t.RT4,
                      _extract_rt_from_call(self.vpn, '_advertise_route', 1))
        self.assertNotIn(t.RT3,
                         _extract_rt_from_call(self.vpn,
                                               '_advertise_route', 1))
        self.assertIn(RTRecord3,
                      _extract_rtrec_from_call(self.vpn,
                                               '_advertise_route', 1))
        self.assertEqual(
            t.RT5,
            _extract_traffic_redirect_from_call(self.vpn,
                                                '_advertise_route', 1))
        # check that second event is for re-advertised route flow_nlri1 and
        #  contains what we expect
        route_entry = self.vpn._advertise_route.call_args_list[1][0][0]
        self.assertNotEqual(flow_nlri1.rd, route_entry.nlri.rd)
        self.assertEqual(self.vpn.instance_rd, route_entry.nlri.rd)
        # dataplane *not* supposed to be updated for this route
        self.mock_dp.setup_dataplane_for_remote_endpoint.assert_not_called()

    def _check_attract_traffic(self, method, redirect_rts,
                               expected_classifiers):
        self.assertEqual(len(expected_classifiers),
                         getattr(self.vpn, method).call_count)

        for index, classifier in enumerate(expected_classifiers):
            if not classifier:
                # Skip advertisement to exported route targets
                if (self.vpn.export_rts == _extract_rt_from_call(
                        self.vpn,
                        method,
                        index)):
                    continue

                # 1 - re-advertisement of a default route supposed to happen
                # to RT4
                self.assertIn(self.vpn.readvertise_to_rts[0],
                              _extract_rt_from_call(self.vpn, method, index))

                ipvpn_nlri = _extract_nlri_from_call(self.vpn, method, index)
                self.assertEqual(ipvpn.DEFAULT_ADDR_PREFIX,
                                 ipvpn_nlri.cidr.prefix())

                if self.vpn.readvertise:
                    self.assertNotIn(self.vpn.readvertise_from_rts[0],
                                     _extract_rt_from_call(self.vpn,
                                                           method, index))
            else:
                # 2 - advertisement of FlowSpec NLRI supposed to happen to RT5
                #     for traffic redirection to RT4 on TCP destination port 80
                flow_nlri = _extract_nlri_from_call(self.vpn, method, index)
                self.assertIsInstance(flow_nlri, exa.Flow)

                self.assertEqual(flow_nlri.rd, self.vpn.instance_rd)
                self.assertIn(redirect_rts[0],
                              _extract_rt_from_call(self.vpn, method, index))
                self.assertEqual(
                    self.vpn.readvertise_to_rts[0],
                    _extract_traffic_redirect_from_call(self.vpn,
                                                        method, index)
                )
                self.assertEqual(
                    classifier,
                    _extract_traffic_classifier_from_call(self.vpn,
                                                          method, index)
                )

    # unit test for IPVPN traffic redirection
    def test_attract_traffic_re_advertisement(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        event1 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1,
                                       [t.RT3], worker_a, t.NH1, 200)

        # Default and FlowSpec routes are supposed to be advertised
        self.assertEqual(2, self.vpn._advertise_route.call_count)
        self.vpn._withdraw_route.assert_not_called()

        ipvpn_nlri = _extract_nlri_from_call(self.vpn, '_advertise_route', 0)
        self.assertIsInstance(ipvpn_nlri, exa.IPVPN)
        self.assertEqual(ipvpn.DEFAULT_ADDR_PREFIX,
                         ipvpn_nlri.cidr.prefix())

        flow_nlri = _extract_nlri_from_call(self.vpn, '_advertise_route', 1)
        self.assertIsInstance(flow_nlri, exa.Flow)

        self._reset_mocks()

        vpn_nlri2 = self._generate_route_nlri(IP_ADDR_PREFIX2)
        event2 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri2,
                                       [t.RT3], worker_a, t.NH1, 200)

        # Only FlowSpec route is supposed to be advertised
        self.vpn._advertise_route.assert_called_once()
        self.vpn._withdraw_route.assert_not_called()

        flow_nlri = _extract_nlri_from_call(self.vpn, '_advertise_route', 0)
        self.assertIsInstance(flow_nlri, exa.Flow)

        self._reset_mocks()

        self._revert_event(event2)

        # Only FlowSpec route is supposed to be withdrawn
        self.vpn._withdraw_route.assert_called_once()
        self.vpn._advertise_route.assert_not_called()

        flow_nlri = _extract_nlri_from_call(self.vpn, '_withdraw_route', 0)
        self.assertIsInstance(flow_nlri, exa.Flow)

        self._reset_mocks()

        self._revert_event(event1)

        # Default and FlowSpec routes are supposed to be withdrawn
        self.assertEqual(2, self.vpn._withdraw_route.call_count)
        self.vpn._advertise_route.assert_not_called()

        ipvpn_nlri = _extract_nlri_from_call(self.vpn, '_withdraw_route', 0)
        self.assertIsInstance(ipvpn_nlri, exa.IPVPN)
        self.assertEqual(ipvpn.DEFAULT_ADDR_PREFIX,
                         ipvpn_nlri.cidr.prefix())

        flow_nlri = _extract_nlri_from_call(self.vpn, '_withdraw_route', 1)
        self.assertIsInstance(flow_nlri, exa.Flow)

    def test_attract_traffic_single_prefix_advertise(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        self._check_attract_traffic('_advertise_route',
                                    ATTRACT_TRAFFIC_1['redirect_rts'],
                                    [None, TC1])

    def test_attract_traffic_single_prefix_withdraw(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        event1 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1,
                                       [t.RT3], worker_a, t.NH1, 200)

        self.assertEqual(2, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        self._revert_event(event1)

        self._check_attract_traffic('_withdraw_route',
                                    ATTRACT_TRAFFIC_1['redirect_rts'],
                                    [None, TC1])

    def test_attract_traffic_multiple_prefix_advertise(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        vpn_nlri2 = self._generate_route_nlri(IP_ADDR_PREFIX2)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri2, [t.RT3],
                              worker_a, t.NH1, 200)

        self._check_attract_traffic(
            '_advertise_route',
            ATTRACT_TRAFFIC_1['redirect_rts'],
            [None, TC1, TC2])

    def test_attract_traffic_multiple_prefix_withdraw(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        event1 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1,
                                       [t.RT3], worker_a, t.NH1, 200)

        vpn_nlri2 = self._generate_route_nlri(IP_ADDR_PREFIX2)
        event2 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri2,
                                       [t.RT3], worker_a, t.NH1, 200)

        self.assertEqual(3, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        self._revert_event(event2)
        self._revert_event(event1)

        self._check_attract_traffic(
            '_withdraw_route',
            ATTRACT_TRAFFIC_1['redirect_rts'],
            [TC2, None, TC1])

    def test_attract_traffic_static_dest_prefix_no_readvertise_advertise(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier and a static destination prefix, to a specific route
        # target
        self._config_vrf_with_attract_traffic(ATTRACT_STATIC_1,
                                              no_readvertise=True)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        self._check_attract_traffic('_advertise_route',
                                    ATTRACT_STATIC_1['redirect_rts'],
                                    [None, None, TC_STATIC1])

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        self.vpn._advertise_route.assert_not_called()

    def test_attract_traffic_static_dest_prefix_advertise(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier and a static destination prefix, to a specific route
        # target
        self._config_vrf_with_attract_traffic(ATTRACT_STATIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        self._check_attract_traffic('_advertise_route',
                                    ATTRACT_STATIC_1['redirect_rts'],
                                    [None, None, TC_STATIC1])

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        self._check_attract_traffic('_advertise_route',
                                    ATTRACT_STATIC_1['redirect_rts'],
                                    [None, TC1])

    def test_attract_traffic_static_dest_prefix_advertise_multiple(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier and multiple static destination prefixes, to a specific
        # route target
        self._config_vrf_with_attract_traffic(ATTRACT_STATIC_2)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        self._check_attract_traffic('_advertise_route',
                                    ATTRACT_STATIC_1['redirect_rts'],
                                    [None, None, TC_STATIC1, None, TC_STATIC2])

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        self._check_attract_traffic('_advertise_route',
                                    ATTRACT_STATIC_1['redirect_rts'],
                                    [None, TC1])

    def test_redirected_vrf_single_flow_advertised(self):
        self._mock_vpnmanager_for_attract_traffic()

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        # FlowSpec route
        flow_nlri1 = self._generate_flow_spec_nlri(TC1)
        self._new_flow_event(engine.RouteEvent.ADVERTISE, flow_nlri1, [t.RT5],
                             [t.RT1], worker_a)

        redirect_rt5 = t._rt_to_string(t.RT5)
        self.manager.redirect_traffic_to_vpn.assert_called_once()
        self.assertIn(TC1,
                      self.vpn.redirect_rt_2_classifiers[redirect_rt5])

    def test_redirected_vrf_multiple_flow_advertised(self):
        self._mock_vpnmanager_for_attract_traffic()

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        # FlowSpec route
        flow_nlri1 = self._generate_flow_spec_nlri(TC1)
        self._new_flow_event(engine.RouteEvent.ADVERTISE, flow_nlri1, [t.RT5],
                             [t.RT1], worker_a)
        flow_nlri2 = self._generate_flow_spec_nlri(TC2)
        self._new_flow_event(engine.RouteEvent.ADVERTISE, flow_nlri2, [t.RT5],
                             [t.RT1], worker_a)

        redirect_rt5 = t._rt_to_string(t.RT5)
        self.assertEqual(2, self.manager.redirect_traffic_to_vpn.call_count)
        self.assertIn(TC1,
                      self.vpn.redirect_rt_2_classifiers[redirect_rt5])
        self.assertIn(TC2,
                      self.vpn.redirect_rt_2_classifiers[redirect_rt5])

    def test_redirected_vrf_multiple_flow_withdrawn(self):
        self._mock_vpnmanager_for_attract_traffic()

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.vpn._advertise_route.assert_called_once()

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        # FlowSpec route

        flow_nlri1 = self._generate_flow_spec_nlri(TC1)
        event1 = self._new_flow_event(engine.RouteEvent.ADVERTISE, flow_nlri1,
                                      [t.RT5], [t.RT1], worker_a)
        flow_nlri2 = self._generate_flow_spec_nlri(TC2)
        event2 = self._new_flow_event(engine.RouteEvent.ADVERTISE, flow_nlri2,
                                      [t.RT5], [t.RT1], worker_a)

        self.assertEqual(2, self.manager.redirect_traffic_to_vpn.call_count)

        self._reset_mocks_vpnmanager()

        self._revert_event(event2)

        redirect_rt5 = t._rt_to_string(t.RT5)
        self.assertNotIn(TC2,
                         self.vpn.redirect_rt_2_classifiers[redirect_rt5])

        self._revert_event(event1)

        self.assertTrue(not self.vpn.redirect_rt_2_classifiers)
        self.manager.stop_redirect_to_vpn.assert_called_once()

    def test_load_balancing_single_prefix_advertise(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.assertEqual(2, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        self._check_attract_traffic('_advertise_route',
                                    ATTRACT_TRAFFIC_1['redirect_rts'],
                                    [None, None, TC1])

    def test_load_balancing_single_prefix_withdraw(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.assertEqual(2, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        event1 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1,
                                       [t.RT3], worker_a, t.NH1, 200)

        self.assertEqual(3, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        self._revert_event(event1)

        self._check_attract_traffic('_withdraw_route',
                                    ATTRACT_TRAFFIC_1['redirect_rts'],
                                    [None, None, TC1])

    def test_load_balancing_multiple_prefix_advertise(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.assertEqual(2, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        vpn_nlri2 = self._generate_route_nlri(IP_ADDR_PREFIX2)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri2, [t.RT3],
                              worker_a, t.NH1, 200)

        self._check_attract_traffic(
            '_advertise_route',
            ATTRACT_TRAFFIC_1['redirect_rts'],
            [None, None, TC1, TC2])

    def test_load_balancing_multiple_prefix_withdraw(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.assertEqual(2, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        event1 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1,
                                       [t.RT3], worker_a, t.NH1, 200)

        vpn_nlri2 = self._generate_route_nlri(IP_ADDR_PREFIX2)
        event2 = self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri2,
                                       [t.RT3], worker_a, t.NH1, 200)

        self.assertEqual(4, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        self._revert_event(event2)
        self._revert_event(event1)

        self._check_attract_traffic(
            '_withdraw_route',
            ATTRACT_TRAFFIC_1['redirect_rts'],
            [TC2, None, None, TC1])

    def test_load_balancing_new_plug(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.assertEqual(2, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        self.assertEqual(3, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        self.vpn.vif_plugged(MAC3, IP3, LOCAL_PORT1)

        self._check_attract_traffic(
            '_advertise_route',
            ATTRACT_TRAFFIC_1['redirect_rts'],
            [None, None])

    def test_load_balancing_unplug_all(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        self.vpn.vif_plugged(MAC1, IP1, LOCAL_PORT1)
        self.vpn.vif_plugged(MAC2, IP2, LOCAL_PORT1)

        # new Route for plugged if supposed to be advertised
        self.assertEqual(2, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1, [t.RT3],
                              worker_a, t.NH1, 200)

        self.assertEqual(3, self.vpn._advertise_route.call_count)

        self._reset_mocks()

        self.vpn.vif_unplugged(MAC1, IP1)

        self._check_attract_traffic(
            '_withdraw_route',
            ATTRACT_TRAFFIC_1['redirect_rts'],
            [None, None])

        self._reset_mocks()

        self.vpn.vif_unplugged(MAC2, IP2)

        self._check_attract_traffic(
            '_withdraw_route',
            ATTRACT_TRAFFIC_1['redirect_rts'],
            [None, TC1, None])

    def test_cleanup_assist(self):
        # Configure VRF to generate traffic redirection, based on a 5-tuple
        # classifier, to a specific route target
        self._config_vrf_with_attract_traffic(ATTRACT_TRAFFIC_1)

        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        vpn_nlri1 = self._generate_route_nlri(IP_ADDR_PREFIX1)
        self._new_route_event(engine.RouteEvent.ADVERTISE, vpn_nlri1,
                              [t.RT1], worker_a, t.NH1, 200)

        # FlowSpec route
        flow_nlri1 = self._generate_flow_spec_nlri(TC1)
        self._new_flow_event(engine.RouteEvent.ADVERTISE, flow_nlri1,
                             [t.RT5], [t.RT1], worker_a)

        self.mock_dp.needs_cleanup_assist.return_value = False

        with mock.patch.object(self.vpn, 'best_route_removed') as mock_brr:
            self.vpn.stop()
            mock_brr.assert_called_once()
            self.assertIsInstance(mock_brr.call_args[0][1].nlri,
                                  flowspec.Flow)



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.fullstack.resources import config as neutron_cfg


class ML2ConfigFixture(neutron_cfg.ML2ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, tenant_network_types):
        super(ML2ConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir, tenant_network_types)

        if env_desc.bagpipe_ml2:
            self.config['ml2']['type_drivers'] = 'vxlan'



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from networking_bagpipe.agent import bagpipe_bgp_agent as agent
from networking_bagpipe.bagpipe_bgp import constants as bbgp_const

from networking_bagpipe.tests.unit.agent.common \
    import constants as const

from neutron.tests import base


def rts_as_set(rts):
    return {k: set(rt_list)
            for k, rt_list in rts.items()}


def _attachments_gen(vpn_type, network, port, rts):
    attachment = {
        'network_id': network['network_id'],
        vpn_type: [dict(
            gateway_ip=network['gateway_ip'],
            **port
        )]
    }
    attachment[vpn_type][0].update(**rts_as_set(rts))
    return attachment


class TestBaGPipeBGPAgentSingleService(base.BaseTestCase):

    def setUp(self):
        super(TestBaGPipeBGPAgentSingleService, self).setUp()

        self.agent = agent.BaGPipeBGPAgent('Linux bridge agent')
        self.agent._send_attach_local_port = mock.Mock()
        self.agent._send_detach_local_port = mock.Mock()

        self.service1 = mock.Mock(name='SERVICE1')
        self.agent.register_build_callback(self.service1.name,
                                           self.service1._build_port_info)

    def _test_port_plug(self, vpn_type, rts):
        attachments = _attachments_gen(vpn_type, const.NETWORK_INFO1,
                                       const.PORT_INFO1, rts)

        self.service1._build_port_info.return_value = attachments

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (vpn_type,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=vpn_type,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(rts)
            ))
        ])

    def test_evpn_port_plug(self):
        self._test_port_plug(bbgp_const.EVPN, const.EVPN_RT1)

    def test_ipvpn_port_plug(self):
        self._test_port_plug(bbgp_const.IPVPN, const.IPVPN_RT100)

    def test_evpn2ipvpn_port_plug(self):
        attachments = _attachments_gen('evpn', const.NETWORK_INFO1,
                                       const.PORT_INFO1, const.EVPN_RT1)
        attachments.update(_attachments_gen('ipvpn', const.NETWORK_INFO1,
                                            const.PORT_INFO1,
                                            const.IPVPN_RT100))

        self.service1._build_port_info.return_value = attachments

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.EVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(const.EVPN_RT1)
            )),
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.IPVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port={
                    bbgp_const.EVPN: {
                        'id': '%s_%s' % (bbgp_const.EVPN,
                                         const.NETWORK_INFO1['network_id'])
                    }
                },
                **rts_as_set(const.IPVPN_RT100)
            ))
        ])

    def test_epvpn_port_plug_updated_local_port(self):
        attachments = _attachments_gen('evpn', const.NETWORK_INFO1,
                                       const.PORT_INFO1, const.EVPN_RT1)
        attachments['evpn'][0]['local_port'] = const.UPDATED_LOCAL_PORT1

        self.service1._build_port_info.return_value = attachments

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.EVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.UPDATED_LOCAL_PORT1,
                **rts_as_set(const.EVPN_RT1)
            ))
        ])

    def _test_port_plug_refresh_without_detach(self, vpn_type, rts):
        attachments = _attachments_gen(vpn_type, const.NETWORK_INFO1,
                                       const.PORT_INFO1, rts)

        detach_info1 = {
            'network_id': const.NETWORK_INFO1['network_id'],
            vpn_type: dict(**const.PORT_INFO1)
        }

        self.service1._build_port_info.return_value = attachments

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (vpn_type,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=vpn_type,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(rts)
            ))
        ])

        self.agent._send_detach_local_port.assert_not_called()

    @testtools.skip("skip until bug 1744344 is resolved")
    def test_evpn_port_plug_refresh_without_detach(self):
        self._test_port_plug_refresh_without_detach(bbgp_const.EVPN,
                                                    const.EVPN_RT1)

    @testtools.skip("skip until bug 1744344 is resolved")
    def test_ipvpn_port_plug_refresh_without_detach(self):
        self._test_port_plug_refresh_without_detach(bbgp_const.IPVPN,
                                                    const.IPVPN_RT100)

    def _test_port_plug_refresh_with_detach(self, vpn_type, rts):
        attachments = _attachments_gen(vpn_type, const.NETWORK_INFO1,
                                       const.PORT_INFO1, rts)

        detach_info1 = {
            'network_id': const.NETWORK_INFO1['network_id'],
            vpn_type: dict(**const.PORT_INFO1)
        }

        self.service1._build_port_info.return_value = attachments

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_not_called()

        self.agent._send_detach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (vpn_type,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=vpn_type,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                local_port=const.PORT_INFO1['local_port']
            ))
        ])

    @testtools.skip("skip until bug 1744344 is resolved")
    def test_evpn_port_plug_refresh_with_detach(self):
        self._test_port_plug_refresh_with_detach(bbgp_const.EVPN,
                                                 const.EVPN_RT1)

    @testtools.skip("skip until bug 1744344 is resolved")
    def test_ipvpn_port_plug_refresh_with_detach(self):
        self._test_port_plug_refresh_with_detach(bbgp_const.IPVPN,
                                                 const.IPVPN_RT100)

    def test_evpn2ipvpn_port_plug_refresh_with_detach(self):
        attachments = _attachments_gen('evpn', const.NETWORK_INFO1,
                                       const.PORT_INFO1, const.EVPN_RT1)

        detach_info1 = {
            'network_id': const.NETWORK_INFO1['network_id'],
            'ipvpn': dict(**const.PORT_INFO1)
        }

        self.service1._build_port_info.return_value = attachments

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.EVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(const.EVPN_RT1)
            ))
        ])

        self.agent._send_detach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.IPVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                local_port={
                    bbgp_const.EVPN: {
                        'id': '%s_%s' % (bbgp_const.EVPN,
                                         const.NETWORK_INFO1['network_id'])
                    }
                }
            ))
        ])


class TestBaGPipeBGPAgentMultipleServices(base.BaseTestCase):

    def setUp(self):
        super(TestBaGPipeBGPAgentMultipleServices, self).setUp()

        self.agent = agent.BaGPipeBGPAgent('Linux bridge agent')
        self.agent._send_attach_local_port = mock.Mock()
        self.agent._send_detach_local_port = mock.Mock()

        self.service1 = mock.Mock(name='SERVICE1')
        self.agent.register_build_callback(self.service1.name,
                                           self.service1._build_port_info)

        self.service2 = mock.Mock(name='SERVICE2')
        self.agent.register_build_callback(self.service2.name,
                                           self.service2._build_port_info)

    def _merge_rts(self, rt1, rt2):
        return {k: set(rt1[k] + rt2[k]) for k in rt1}

    def test_evpns_port_plug(self):
        attachments_1 = _attachments_gen('evpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.EVPN_RT1)
        self.service1._build_port_info.return_value = attachments_1

        attachments_2 = _attachments_gen('evpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.EVPN_RT2)
        self.service2._build_port_info.return_value = attachments_2

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.EVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **self._merge_rts(const.EVPN_RT1, const.EVPN_RT2)
            ))
        ])

    def test_ipvpns_port_plug(self):
        attachments_1 = _attachments_gen('ipvpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.IPVPN_RT100)
        self.service1._build_port_info.return_value = attachments_1

        attachments_2 = _attachments_gen('ipvpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.IPVPN_RT200)
        self.service2._build_port_info.return_value = attachments_2

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.IPVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **self._merge_rts(const.IPVPN_RT100, const.IPVPN_RT200)
            ))
        ])

    def test_evpn2ipvpn_port_plug(self):
        attachments_1 = _attachments_gen('evpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.EVPN_RT1)
        self.service1._build_port_info.return_value = attachments_1

        attachments_2 = _attachments_gen('ipvpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.IPVPN_RT100)
        self.service2._build_port_info.return_value = attachments_2

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.EVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(const.EVPN_RT1)
            )),
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.IPVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port={
                    bbgp_const.EVPN: {
                        'id': '%s_%s' % (bbgp_const.EVPN,
                                         const.NETWORK_INFO1['network_id'])
                    }
                },
                **rts_as_set(const.IPVPN_RT100)
            ))
        ])

    def test_evpns_port_plug_refresh_without_detach(self):
        attachments_1 = _attachments_gen('evpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.EVPN_RT1)
        self.service1._build_port_info.return_value = attachments_1

        self.service2._build_port_info.return_value = {}

        detach_info1 = {
            'network_id': const.NETWORK_INFO1['network_id'],
            bbgp_const.EVPN: dict(**const.PORT_INFO1)
        }

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.EVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(const.EVPN_RT1)
            ))
        ])

    def test_ipvpns_port_plug_refresh_without_detach(self):
        attachments_1 = _attachments_gen('ipvpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.IPVPN_RT100)
        self.service1._build_port_info.return_value = attachments_1

        self.service2._build_port_info.return_value = {}

        detach_info1 = {
            'network_id': const.NETWORK_INFO1['network_id'],
            bbgp_const.IPVPN: dict(**const.PORT_INFO1)
        }

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.IPVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(const.IPVPN_RT100)
            ))
        ])

    def test_evpn2ipvpn_port_plug_refresh_with_detach(self):
        attachments_1 = _attachments_gen('evpn', const.NETWORK_INFO1,
                                         const.PORT_INFO1, const.EVPN_RT1)
        self.service1._build_port_info.return_value = attachments_1

        self.service2._build_port_info.return_value = {}

        detach_info1 = {
            'network_id': const.NETWORK_INFO1['network_id'],
            bbgp_const.IPVPN: dict(**const.PORT_INFO1)
        }

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.EVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts_as_set(const.EVPN_RT1)
            ))
        ])

        self.agent._send_detach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (bbgp_const.IPVPN,
                                           const.NETWORK_INFO1['network_id']),
                vpn_type=bbgp_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                local_port={
                    bbgp_const.EVPN: {
                        'id': '%s_%s' % (bbgp_const.EVPN,
                                         const.NETWORK_INFO1['network_id'])
                    }
                }
            ))
        ])



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2018 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging

from networking_bagpipe.bagpipe_bgp.common import dataplane_utils
from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp import constants as consts
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.bagpipe_bgp.vpn import evpn

from neutron.plugins.ml2.drivers.openvswitch.agent.common import \
    constants as ovs_const
from neutron.plugins.ml2.drivers.openvswitch.agent.openflow.ovs_ofctl import \
    br_tun
from neutron.plugins.ml2.drivers.openvswitch.agent import ovs_neutron_agent

from neutron_lib import constants as n_consts

LOG = logging.getLogger(__name__)

FLOOD = "flood"
FLOW_PRIORITY = 5


class OVSEVIDataplane(evpn.VPNInstanceDataplane):

    def __init__(self, *args, **kwargs):
        super(OVSEVIDataplane, self).__init__(*args, **kwargs)

        self.bridge = self.driver.bridge
        # OpenFlow 1.3 is needed for mod_vlan_vid
        self.bridge.use_at_least_protocol(ovs_const.OPENFLOW13)

        self.tunnel_mgr = dataplane_utils.ObjectLifecycleManagerProxy(
            self.driver.tunnel_mgr,
            self.instance_id
        )
        self.flooding_ports = set()
        self.vlan = None
        self.local_ip = self.driver.get_local_address()

    def cleanup(self):
        self.bridge.delete_flows(strict=True,
                                 table=ovs_const.FLOOD_TO_TUN,
                                 priority=FLOW_PRIORITY,
                                 dl_vlan=self.vlan)
        self.bridge.delete_group(group_id=self.vlan)

    @log_decorator.log_info
    def vif_plugged(self, mac_address, ip_address_prefix, localport, label,
                    direction):
        if 'vlan' not in localport:
            raise Exception("missing localport['vlan'] parameter")
        if self.vlan and localport['vlan'] != self.vlan:
            raise Exception("inconsistent vlan")
        else:
            self.vlan = localport['vlan']

        # map traffic to this EVI VNI to the right table, similarly as in
        # OVSTunnelBridge.provision_local_vlan
        self.bridge.add_flow(table=ovs_const.VXLAN_TUN_TO_LV,
                             priority=FLOW_PRIORITY,
                             tun_id=self.instance_label,
                             actions=("push_vlan:0x8100,mod_vlan_vid:%d,"
                                      "resubmit(,%s)" %
                                      (self.vlan, ovs_const.LEARN_FROM_TUN)))

    @log_decorator.log_info
    def vif_unplugged(self, mac_address, ip_address_prefix, localport, label,
                      direction, last_endpoint=True):
        self.log.debug("nothing to do on unplug")

    def _local_vni_actions(self, vni):
        # "load:0->NXM_OF_IN_PORT[]" allows the packets coming from br-int
        # via patch-port to go back to br-int via the same port
        return "load:0->NXM_OF_IN_PORT[],set_tunnel:%d,resubmit(,%s)" % (
            vni, ovs_const.VXLAN_TUN_TO_LV)

    @log_decorator.log_info
    def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, vni, nlri,
                                            encaps):
        mac = prefix
        ip = nlri.ip

        # what is done here is similar as
        # OVSTunnelBridge.install_unicast_to_tun, but with local delivery to
        # table VXLAN_TUN_TO_LV for routes advertized locally
        if remote_pe == self.local_ip:
            actions = self._local_vni_actions(vni)
        else:
            port = self.tunnel_mgr.get_object(remote_pe, (vni, mac))
            actions = "set_tunnel:%d,output:%s" % (vni, port)
        self.bridge.add_flow(table=ovs_const.UCAST_TO_TUN,
                             priority=FLOW_PRIORITY,
                             dl_vlan=self.vlan,
                             dl_dst=mac,
                             actions="strip_vlan,%s" % actions)

        # add ARP responder
        if ip:
            self.bridge.install_arp_responder(self.vlan, str(ip), str(mac))

    @log_decorator.log
    def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, vni,
                                             nlri):
        mac = prefix
        ip = nlri.ip

        self.bridge.delete_unicast_to_tun(self.vlan, mac)

        if remote_pe != self.local_ip:
            self.tunnel_mgr.free_object(remote_pe, (vni, mac))

        # cleanup ARP responder
        if ip:
            self.bridge.delete_arp_responder(self.vlan, str(ip))

    @log_decorator.log_info
    def add_dataplane_for_bum_endpoint(self, remote_pe, vni, nlri, encaps):

        if remote_pe == self.local_ip:
            port = "local"
        else:
            port = self.tunnel_mgr.get_object(remote_pe, (vni, FLOOD))
        self.flooding_ports.add((port, vni))

        self._update_flooding_buckets()

    @log_decorator.log_info
    def remove_dataplane_for_bum_endpoint(self, remote_pe, vni, nlri):

        if remote_pe == self.local_ip:
            port = "local"
        else:
            port = self.tunnel_mgr.get_object(remote_pe)

        if port:
            self.flooding_ports.remove((port, vni))

            self._update_flooding_buckets()

            if remote_pe != self.local_ip:
                self.tunnel_mgr.free_object(remote_pe, (vni, FLOOD))

    def _update_flooding_buckets(self):
        buckets = []
        for port, vni in self.flooding_ports:
            if port == "local":
                buckets.append("bucket=strip_vlan,%s" %
                               self._local_vni_actions(vni))
            else:
                buckets.append("bucket=strip_vlan,set_tunnel:%d,output:%s" %
                               (vni, port))
        self.bridge.mod_group(group_id=self.vlan,
                              type='all',
                              buckets=','.join(buckets))
        self.bridge.add_flow(table=ovs_const.FLOOD_TO_TUN,
                             priority=FLOW_PRIORITY,
                             dl_vlan=self.vlan,
                             actions="group:%d" % self.vlan)
        self.log.debug("buckets: %s", buckets)

    def set_gateway_port(self, linuxif, gateway_ip):
        # nothing to do, because we make the assumption that the
        # IPVPN driver is 'ovs' as well, and setup in conjunction
        # with Neutron OVS BGPVPN extension which does the plugging
        # between L2 and L3
        pass

    def gateway_port_down(self, linuxif):
        pass

    # Looking glass ####

    def get_lg_local_info(self, path_prefix):
        return {
            "vlan": self.vlan,
            "flooding-ports": [{"port": str(port), "vni": vni}
                               for port, vni in self.flooding_ports]
        }


class TunnelManager(dataplane_utils.ObjectLifecycleManager):

    def __init__(self, bridge, local_ip):
        super(TunnelManager, self).__init__()

        self.bridge = bridge
        self.local_ip = local_ip

    @log_decorator.log_info
    def create_object(self, remote_ip):
        port_name = ovs_neutron_agent.OVSNeutronAgent.get_tunnel_name(
            n_consts.TYPE_VXLAN, self.local_ip, remote_ip)
        tunnel = self.bridge.add_tunnel_port(port_name,
                                             remote_ip,
                                             self.local_ip,
                                             n_consts.TYPE_VXLAN)

        self.bridge.setup_tunnel_port(n_consts.TYPE_VXLAN, tunnel)
        LOG.debug("tunnel for %s: %s (%s)", remote_ip, port_name, tunnel)

        return tunnel

    @log_decorator.log_info
    def delete_object(self, tunnel):
        self.bridge.delete_port(tunnel)


class OVSDataplaneDriver(dp_drivers.DataplaneDriver):

    dataplane_instance_class = OVSEVIDataplane
    type = consts.EVPN
    ecmp_support = False
    encaps = [exa.Encapsulation(exa.Encapsulation.Type.VXLAN)]

    driver_opts = [
        cfg.StrOpt("ovs_bridge", default="br-tun",
                   help=("Name of the OVS bridge to use, this has to be the "
                         "same as the tunneling bridge of the Neutron OVS "
                         "agent, usually br-tun")),
    ]

    def __init__(self, *args, **kwargs):
        super(OVSDataplaneDriver, self).__init__(*args, **kwargs)

        # copy bagpipe-bgp root helper configuration into neutron's config, so
        # that neutron classes find the right configuration to execute commands
        cfg.CONF.set_default('root_helper',
                             cfg.CONF.COMMON.root_helper,
                             group="AGENT")
        cfg.CONF.set_default('root_helper_daemon',
                             cfg.CONF.COMMON.root_helper_daemon,
                             group="AGENT")

        self.bridge = dataplane_utils.OVSBridgeWithGroups(
            br_tun.OVSTunnelBridge(self.config.ovs_bridge)
        )
        self.tunnel_mgr = TunnelManager(self.bridge,
                                        self.get_local_address())

    def needs_cleanup_assist(self):
        return True

    def reset_state(self):
        # cleanup is taken care of by OVS Neutron Agent
        pass

    # Looking glass ####

    def get_lg_local_info(self, path_prefix):
        return {
            "tunnels": self.tunnel_mgr.infos(),
        }



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path

import fixtures
from oslo_serialization import jsonutils


class JsonFileFixture(fixtures.Fixture):
    """A fixture that knows how to translate configuration to JSON file.

    :param base_filename: the filename to use on disk.
    :param config: a dictionary.
    :param temp_dir: an existing temporary directory to use for storage.
    """

    def __init__(self, base_filename, config, temp_dir):
        super(JsonFileFixture, self).__init__()
        self.base_filename = base_filename
        self.config = config
        self.temp_dir = temp_dir

    def _setUp(self):
        # Need to randomly generate a unique folder to put the file in
        self.filename = os.path.join(self.temp_dir, self.base_filename)
        with open(self.filename, 'w') as f:
            jsonutils.dump(self.config, f, indent=4)
            f.flush()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import bgp_manager
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.tests.unit.bagpipe_bgp import base


class TestRouteTableManager(base.TestCase):

    def setUp(self):
        super(TestRouteTableManager, self).setUp()

        self.bgp_manager = bgp_manager.Manager()

    def test1(self):
        subscription = engine.Subscription(engine.Subscription.ANY_AFI,
                                           engine.Subscription.ANY_SAFI,
                                           engine.Subscription.ANY_RT)

        route_entry = self.bgp_manager._subscription_2_rtc_route_entry(
            subscription)

        self.assertEqual(route_entry.safi, exa.SAFI.rtc, "wrong RTC route")



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce

from oslo_config import cfg
from oslo_log import log as logging

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import bgp_peer_worker
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import exabgp_peer_worker
from networking_bagpipe.bagpipe_bgp.engine import route_table_manager as rtm


LOG = logging.getLogger(__name__)

# SAFIs for which RFC4684 is effective
RTC_SAFIS = (exa.SAFI.mpls_vpn, exa.SAFI.evpn)


class Manager(engine.EventSource, lg.LookingGlassMixin):

    _instance = None

    def __init__(self):

        LOG.debug("Instantiating BGPManager")

        if cfg.CONF.BGP.enable_rtc:
            first_local_subscriber_callback = self.rtc_advertisement_for_sub
            last_local_subscriber_callback = self.rtc_withdrawal_for_sub
        else:
            first_local_subscriber_callback = None
            last_local_subscriber_callback = None

        self.rtm = rtm.RouteTableManager(first_local_subscriber_callback,
                                         last_local_subscriber_callback)

        self.rtm.start()

        self.peers = {}
        if cfg.CONF.BGP.peers:
            for peer_address in cfg.CONF.BGP.peers:
                LOG.debug("Creating a peer worker for %s", peer_address)
                peer_worker = exabgp_peer_worker.ExaBGPPeerWorker(self,
                                                                  peer_address)
                self.peers[peer_address] = peer_worker
                peer_worker.start()

        # we need a .name since we'll masquerade as a route_entry source
        self.name = "BGPManager"

        engine.EventSource.__init__(self, self.rtm)

    def __repr__(self):
        return self.__class__.__name__

    @log_decorator.log
    def stop(self):
        for peer in self.peers.values():
            peer.stop()
        self.rtm.stop()
        for peer in self.peers.values():
            peer.join()
        self.rtm.join()

    def get_local_address(self):
        return cfg.CONF.BGP.local_address

    @log_decorator.log
    def rtc_advertisement_for_sub(self, sub):
        if sub.safi in RTC_SAFIS:
            event = engine.RouteEvent(
                engine.RouteEvent.ADVERTISE,
                self._subscription_2_rtc_route_entry(sub),
                self)
            LOG.debug("Based on subscription => synthesized RTC %s", event)
            self.rtm.enqueue(event)

    @log_decorator.log
    def rtc_withdrawal_for_sub(self, sub):
        if sub.safi in RTC_SAFIS:
            event = engine.RouteEvent(
                engine.RouteEvent.WITHDRAW,
                self._subscription_2_rtc_route_entry(sub),
                self)
            LOG.debug("Based on unsubscription => synthesized withdraw"
                      " for RTC %s", event)
            self.rtm.enqueue(event)

    def _subscription_2_rtc_route_entry(self, subscription):

        nlri = exa.RTC.new(exa.AFI.ipv4,
                           exa.SAFI.rtc,
                           cfg.CONF.BGP.my_as,
                           subscription.route_target,
                           exa.IP.create(self.get_local_address()))

        route_entry = engine.RouteEntry(nlri)

        return route_entry

    @classmethod
    @utils.oslo_synchronized('BGPManager')
    def _create_instance(cls):
        if not cls.has_instance():
            cls._instance = cls()

    @classmethod
    def has_instance(cls):
        return cls._instance is not None

    @classmethod
    def clear_instance(cls):
        cls._instance = None

    @classmethod
    def get_instance(cls):
        # double checked locking
        if not cls.has_instance():
            cls._create_instance()
        return cls._instance

    # Looking Glass Functions ###################

    def get_lg_map(self):
        return {"peers":   (lg.COLLECTION, (self.get_lg_peer_list,
                                            self.get_lg_peer_path_item)),
                "routes":  (lg.FORWARD, self.rtm),
                "workers": (lg.FORWARD, self.rtm), }

    def get_established_peers_count(self):
        return reduce(lambda count, peer: count +
                      (isinstance(peer, bgp_peer_worker.BGPPeerWorker) and
                       peer.is_established()),
                      self.peers.values(), 0)

    def get_lg_peer_list(self):
        return [{"id": peer.peer_address,
                 "state": peer.fsm.state} for peer in self.peers.values()]

    def get_lg_peer_path_item(self, path_item):
        return self.peers[path_item]



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron.tests.unit.plugins.ml2 import test_plugin

from neutron_lib import constants as n_consts


class TestBaGpipeML2MechDriver(test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['bagpipe']

    def setUp(self):
        cfg.CONF.set_override('type_drivers',
                              n_consts.TYPE_VXLAN,
                              'ml2')
        cfg.CONF.set_override('tenant_network_types',
                              n_consts.TYPE_VXLAN,
                              'ml2')

        super(TestBaGpipeML2MechDriver, self).setUp()

    def test_setup(self):
        pass



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import eventlet
eventlet.monkey_patch()

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_config import types
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_bagpipe.agent import agent_base_info
from networking_bagpipe.agent import bagpipe_bgp_agent
from networking_bagpipe.bagpipe_bgp import constants as bbgp_const

from neutron.agent.linux import ip_lib
from neutron.common import config as common_config
from neutron.plugins.ml2.drivers.linuxbridge.agent import \
    linuxbridge_neutron_agent as lnx_agt

from neutron_lib.agent import l2_extension
from neutron_lib import constants as n_const


LOG = logging.getLogger(__name__)

BAGPIPE_L2_SERVICE = 'bagpipe_l2'

opts = [
    cfg.ListOpt('as_number', default=[64512],
                item_type=types.Integer(min=1, max=2**32),
                help=("Autonomous System number used to generate BGP RTs for"
                      "E-VPNs used by bagpipe ML2 (more than one is possible,"
                      "to allow a deployment to do a 2-step transition "
                      "to change the AS number used)")
                )
]
cfg.CONF.register_opts(opts, "ml2_bagpipe_extension")


class BagpipeML2AgentExtension(l2_extension.L2AgentExtension,
                               agent_base_info.BaseInfoManager):

    def initialize(self, connection, driver_type):

        self.bagpipe_bgp_agent = (
            bagpipe_bgp_agent.BaGPipeBGPAgent.get_instance(
                n_const.AGENT_TYPE_LINUXBRIDGE)
        )

        self.bagpipe_bgp_agent.register_build_callback(
            BAGPIPE_L2_SERVICE,
            self.build_bagpipe_l2_attach_info)

        self.ports = set()
        self.bagpipe_bgp_agent.register_port_list(BAGPIPE_L2_SERVICE,
                                                  self.ports)

    @log_helpers.log_method_call
    def build_bagpipe_l2_attach_info(self, port_id):
        port_info = self.ports_info.get(port_id)

        if not port_info:
            LOG.debug("no info for port %s", port_id)
            return {}

        LOG.debug("segmentation id: %s", port_info.network.segmentation_id)

        as_numbers = cfg.CONF.ml2_bagpipe_extension.as_number
        bagpipe_rts = [
            "%s:%s" % (as_number, port_info.network.segmentation_id)
            for as_number in as_numbers
        ]

        attach_info = self._base_attach_info(port_info)
        attach_info.update({
            'linuxbr': lnx_agt.LinuxBridgeManager.get_bridge_name(
                port_info.network.id
            ),
            'vni': port_info.network.segmentation_id,
            bbgp_const.RT_IMPORT: bagpipe_rts,
            bbgp_const.RT_EXPORT: bagpipe_rts
        })

        return {
            'network_id': port_info.network.id,
            bbgp_const.EVPN: [
                attach_info
            ]
        }

    def _base_attach_info(self, port_info):
        info = {
            'mac_address': port_info.mac_address,
            'local_port': {
                'linuxif': lnx_agt.LinuxBridgeManager.get_tap_device_name(
                    port_info.id)
            }
        }
        if port_info.ip_address:
            info.update({'ip_address': port_info.ip_address})
        return info

    @lockutils.synchronized('bagpipe-ml2')
    @log_helpers.log_method_call
    def handle_port(self, context, data):
        if data.get('network_type') != n_const.TYPE_VXLAN:
            LOG.debug("network is not of type vxlan, not handled by this "
                      "extension")
            return

        port_id = data['port_id']

        tap_device_name = lnx_agt.LinuxBridgeManager.get_tap_device_name(
            port_id)
        if not ip_lib.device_exists(tap_device_name):
            LOG.debug('skip non-existing port %s', port_id)
            return

        net_id = data['network_id']
        net_info, port_info = (
            self._get_network_port_infos(net_id, port_id)
        )

        def delete_hook():
            self._delete_port(context, {'port_id': port_info.id})

        port_info.update_admin_state(data, delete_hook)
        if not port_info.admin_state_up:
            return

        port_info.mac_address = data['mac_address']

        # take the first IPv4 (error if none, warning if many)
        ip_address = None
        for alloc in data.get('fixed_ips'):
            if '.' in alloc['ip_address']:
                if not ip_address:
                    ip_address = alloc['ip_address']
                else:
                    LOG.warning("multiple IPv4 addresses for %s, ignoring %s",
                                port_id, alloc['ip_address'])
        if ip_address is None:
            LOG.debug("no IP address for port %s", port_id)

        port_info.ip_address = ip_address

        net_info.segmentation_id = data['segmentation_id']

        self.bagpipe_bgp_agent.do_port_plug(port_id)

        self.ports.add(port_id)

    @lockutils.synchronized('bagpipe-ml2')
    def delete_port(self, context, data):
        self._delete_port(context, data)

    # un-synchronized version, to be called indirectly from handle_port
    @log_helpers.log_method_call
    def _delete_port(self, context, data):
        port_id = data['port_id']
        port_info = self.ports_info.get(port_id)

        if port_info:
            detach_info = {
                'network_id': port_info.network.id,
                bbgp_const.EVPN: self._base_attach_info(port_info)
            }

            self._remove_network_port_infos(port_info.network.id, port_id)
            self.ports.remove(port_id)

            self.bagpipe_bgp_agent.do_port_plug_refresh(port_id,
                                                        detach_info)


def main():
    common_config.init(sys.argv[1:])
    common_config.setup_logging()
    LOG.warning('This modified agent is not needed anymore. The normal '
                'neutron linuxbridge agent should be used instead, along with'
                'networks of type VXLAN, rather than RT.')



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

from oslo_config import cfg
from oslo_utils import uuidutils

from neutron_lib.api.definitions import provider_net as provider

from neutron.tests.unit.db import test_db_base_plugin_v2 as test_db_plugin

SRC_CIDR = '10.10.0.0/24'
SRC_GATEWAY = '10.10.0.1'

DEST_CIDR = '10.100.0.0/24'
DEST_GATEWAY = '10.100.0.1'


class NeutronDbPluginV2TestCase(test_db_plugin.NeutronDbPluginV2TestCase):
    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):

        if not plugin:
            plugin = 'neutron.plugins.ml2.plugin.Ml2Plugin'

        cfg.CONF.set_override('tenant_network_types', ['vxlan'], group='ml2')
        cfg.CONF.set_override(
            'vni_ranges', ['1:1000'], group='ml2_type_vxlan')
        cfg.CONF.set_override(
            'mechanism_drivers', ['linuxbridge'], group='ml2')

        super(NeutronDbPluginV2TestCase, self).setUp(
            ext_mgr=ext_mgr,
            plugin=plugin,
            service_plugins=service_plugins
        )

        self._tenant_id = uuidutils.generate_uuid()
        self._network = self._make_network(
            self.fmt, 'test_net',
            True)
        self._subnet = self._make_subnet(
            self.fmt, self._network, gateway='10.0.0.1',
            cidr='10.0.0.0/24', ip_version=4
        )

        self._src_net = self._make_network(
            self.fmt, 'src_net',
            True)
        self._src_subnet = self._make_subnet(
            self.fmt, self._src_net, gateway=SRC_GATEWAY,
            cidr=SRC_CIDR, ip_version=4
        )
        self._src_rt = (
            ':'.join(['64512',
                      str(self._src_net['network'][provider.SEGMENTATION_ID])
                      ])
        )

        self._dest_net = self._make_network(
            self.fmt, 'dest_net',
            True)
        self._dest_subnet = self._make_subnet(
            self.fmt, self._dest_net, gateway=DEST_GATEWAY,
            cidr=DEST_CIDR, ip_version=4
        )
        self._dest_rt = (
            ':'.join(['64512',
                      str(self._dest_net['network'][provider.SEGMENTATION_ID])
                      ])
        )

    def tearDown(self):
        super(NeutronDbPluginV2TestCase, self).tearDown()

    @contextlib.contextmanager
    def port(self, fmt=None, **kwargs):
        net_id = kwargs.get('network_id', self._network['network']['id'])
        port = self._make_port(fmt or self.fmt, net_id, **kwargs)
        yield port



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from oslo_utils import uuidutils

from neutron_lib import context

from neutron.objects import network
from neutron.objects import ports

from networking_bagpipe.objects import sfc as sfc_obj

from networking_sfc.db import sfc_db

from neutron.tests.unit.objects import test_base
from neutron.tests.unit import testlib_api

INGRESS_GW_IP = "10.10.0.1"
INGRESS_MAC = "de:ad:00:00:be:ef"

EGRESS_GW_IP = "10.100.0.1"
EGRESS_MAC = "ba:be:00:00:b0:0b"

RT1 = '64512:7'
RT2 = '64512:42'


class _BaGPipeObjectsTestCommon(object):

    def _create_test_chain_hop(self, rt, ingress_network=None,
                               egress_network=None):
        chainhop_obj = sfc_obj.BaGPipeChainHop(
            self.context,
            id=uuidutils.generate_uuid(),
            portchain_id=uuidutils.generate_uuid(),
            rts=[rt],
            ingress_network=ingress_network,
            egress_network=egress_network,
            ingress_gw=INGRESS_GW_IP,
            egress_gw=EGRESS_GW_IP)

        chainhop_obj.create()

        return chainhop_obj

    def _create_test_port_chain(self, chain_id=0):
        portchain_db = sfc_db.PortChain(id=uuidutils.generate_uuid(),
                                        project_id=self.context.tenant_id,
                                        chain_id=chain_id)

        self.context.session.add(portchain_db)

        return portchain_db


class BaGPipeChainHopDbObjectTestCase(test_base.BaseDbObjectTestCase,
                                      testlib_api.SqlTestCase,
                                      _BaGPipeObjectsTestCommon):

    _test_class = sfc_obj.BaGPipeChainHop

    def setUp(self):
        test_base.BaseDbObjectTestCase.setUp(self)

        self._group_id = 0
        self._chain_id = 0

        self.update_obj_fields({
            'ingress_network': self._create_test_network_id(),
            'egress_network': self._create_test_network_id(),
            'ingress_ppg': self._create_test_ppg_id(),
            'egress_ppg': self._create_test_ppg_id(),
            'portchain_id': self._create_test_pc_id()
        })

    def _create_test_ppg(self):
        test_ppg = sfc_db.PortPairGroup(id=uuidutils.generate_uuid(),
                                        project_id=self.context.tenant_id,
                                        group_id=self._group_id)

        self.context.session.add(test_ppg)
        self._group_id += 1

        return test_ppg

    def _create_test_ppg_id(self):
        return self._create_test_ppg().id

    def _create_test_pc(self):
        test_pc = sfc_db.PortChain(id=uuidutils.generate_uuid(),
                                   project_id=self.context.tenant_id,
                                   chain_id=self._chain_id)

        self.context.session.add(test_pc)
        self._chain_id += 1

        return test_pc

    def _create_test_pc_id(self):
        return self._create_test_pc().id

    def test_create_single_transaction(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_update_single_transaction(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_get_object_single_transaction(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_get_objects_single_transaction(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_update_objects(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_get_objects_queries_constant(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_get_objects_by_port_id(self):
        ingress_network = self._create_test_network()
        ingress_port = self._create_test_port(network_id=ingress_network.id)
        egress_network = self._create_test_network()

        chain_hop1 = self._create_test_chain_hop(RT1,
                                                 ingress_network.id,
                                                 egress_network.id)

        chain_hop2 = self._create_test_chain_hop(RT2,
                                                 egress_network.id,
                                                 ingress_network.id)

        port_hops = sfc_obj.BaGPipeChainHop.get_objects(
            self.context,
            port_id=ingress_port.id)

        self.assertEqual(2, len(port_hops))
        self.assertDictContainsSubset(chain_hop1.to_dict(),
                                      port_hops[0].to_dict())
        self.assertIn(ingress_port.id,
                      port_hops[0].ingress_ports)

        self.assertDictContainsSubset(chain_hop2.to_dict(),
                                      port_hops[1].to_dict())
        self.assertIn(ingress_port.id,
                      port_hops[1].egress_ports)


class BaGPipePortHopsObjectTestCase(testlib_api.SqlTestCase,
                                    _BaGPipeObjectsTestCommon):

    def setUp(self):
        super(BaGPipePortHopsObjectTestCase, self).setUp()
        self.context = context.get_admin_context()

        self.ingress_network = network.Network(self.context)
        self.ingress_network.create()

        self.ingress_port = ports.Port(self.context,
                                       network_id=self.ingress_network.id,
                                       mac_address=netaddr.EUI(
                                           INGRESS_MAC,
                                           dialect=netaddr.mac_unix_expanded),
                                       device_id='test_device_id',
                                       device_owner='compute:None',
                                       status="DUMMY_STATUS",
                                       admin_state_up=True)
        self.ingress_port.create()

        self.egress_network = network.Network(self.context)
        self.egress_network.create()

        self.egress_port = ports.Port(self.context,
                                      network_id=self.egress_network.id,
                                      mac_address=netaddr.EUI(
                                          EGRESS_MAC,
                                          dialect=netaddr.mac_unix_expanded),
                                      device_id='test_device_id',
                                      device_owner='compute:None',
                                      status="DUMMY_STATUS",
                                      admin_state_up=True)
        self.egress_port.create()

        self.port_chain1 = self._create_test_port_chain()

        self.chain_hop1 = (
            self._create_test_chain_hop(RT1,
                                        ingress_network=self.egress_network.id)
        )

        self.chain_hop2 = (
            self._create_test_chain_hop(RT2,
                                        egress_network=self.ingress_network.id)
        )

    def _create_test_port_chain(self):
        portchain_db = sfc_db.PortChain(id=uuidutils.generate_uuid(),
                                        project_id=self.context.tenant_id,
                                        chain_id=0)

        self.context.session.add(portchain_db)

        return portchain_db

    def test_init(self):
        sfc_obj.BaGPipePortHops(
            port_id=self.ingress_port.id,
            ingress_hops=[self.chain_hop1],
            egress_hops=[self.chain_hop2])

    def test_get_object_ingress_chain_hop(self):
        port_hops = sfc_obj.BaGPipePortHops.get_object(
            self.context,
            port_id=self.ingress_port.id
        )

        self.assertEqual(0, len(port_hops.ingress_hops))
        self.assertEqual(1, len(port_hops.egress_hops))
        self.assertDictContainsSubset(self.chain_hop2.to_dict(),
                                      port_hops.egress_hops[0].to_dict())
        self.assertEqual(self.ingress_port.id,
                         port_hops.port_id)

    def test_get_object_egress_chain_hop(self):
        port_hops = sfc_obj.BaGPipePortHops.get_object(
            self.context,
            port_id=self.egress_port.id
        )

        self.assertEqual(1, len(port_hops.ingress_hops))
        self.assertEqual(0, len(port_hops.egress_hops))
        self.assertDictContainsSubset(self.chain_hop1.to_dict(),
                                      port_hops.ingress_hops[0].to_dict())
        self.assertEqual(self.egress_port.id,
                         port_hops.port_id)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from distutils import version  # pylint: disable=no-name-in-module
from oslo_config import cfg
from oslo_log import log as logging
import six
import stevedore

from networking_bagpipe.bagpipe_bgp.common import config
from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import run_command
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import constants
from networking_bagpipe.bagpipe_bgp.engine import exa


LOG = logging.getLogger(__name__)


# NOTE(tmorin): have dataplane_local_address default to
#               cfg.CONF.BGP.local_address does not work (import order issue)
# TODO(tmorin): list possible values for dataplane_driver,
#               see what neutron-db-manage does
dataplane_common_opts = [
    cfg.Opt("dataplane_local_address",
            type=config.InterfaceAddress(),
            help=("IP address to use as next-hop in our route "
                  "advertisements, will be used to send us "
                  "VPN traffic")),
    cfg.StrOpt("dataplane_driver", default="dummy",
               help="Dataplane driver.")
]


for vpn_type in constants.VPN_TYPES:
    cfg.CONF.register_opts(dataplane_common_opts,
                           constants.config_group(vpn_type))


def register_driver_opts(vpn_type, driver_opts):
    cfg.CONF.register_opts(driver_opts, constants.config_group(vpn_type))

# prefix for setuptools entry points for dataplane drivers
DATAPLANE_DRIVER_ENTRY_POINT_PFX = "bagpipe.dataplane"


def instantiate_dataplane_drivers():
    LOG.debug("Building dataplane drivers...")

    if 'DATAPLANE_DRIVER' in cfg.CONF:
        LOG.warning("Config file is obsolete, should have a "
                    "DATAPLANE_DRIVER_IPVPN section instead of"
                    " DATAPLANE_DRIVER")

    drivers = {}
    for vpn_type in constants.VPN_TYPES:
        dp_config = cfg.CONF.get(constants.config_group(vpn_type))

        driver_name = dp_config.dataplane_driver
        LOG.debug("Instantiating dataplane driver for %s, with %s",
                  vpn_type, driver_name)
        try:
            driver_class = stevedore.driver.DriverManager(
                namespace='%s.%s' % (DATAPLANE_DRIVER_ENTRY_POINT_PFX,
                                     vpn_type),
                name=driver_name,
            ).driver

            drivers[vpn_type] = driver_class()
        except Exception:
            LOG.exception("Error while instantiating dataplane driver for "
                          "%s with %s", vpn_type, driver_name)
            raise

    return drivers


@six.add_metaclass(abc.ABCMeta)
class DataplaneDriver(lg.LookingGlassLocalLogger):
    '''Dataplane driver

    The initialisation workflow is the following:
    - on startup, the selected dataplane driver is loaded (its __init__ is
      called)
    - on the first attachment of an interface that matches the driver's type
      the following happens:
      * the driver reset_state() method is called
      * the driver initialize() method is called

    The cleanup workflow (what happens when bagpipe-bgp-cleanup is called) is
    the following:
    - the dataplane driver is loaded (its __init__ is called)
    - reset_state() method is called

    As a consequence, whether some init action should go in __init__ or
    initialize depends on what reset_state does:
    - __init__ should not do things that reset_state will revert
    - initialize can do things that reset_state will revert
    '''

    type = None

    dataplane_instance_class = object  # has to be overridden by subclasses
    encaps = [exa.Encapsulation(exa.Encapsulation.Type.DEFAULT),
              exa.Encapsulation(exa.Encapsulation.Type.MPLS)]
    ecmp_support = False
    required_kernel = None

    driver_opts = []

    @log_decorator.log
    def __init__(self):
        lg.LookingGlassLocalLogger.__init__(self)

        cfg.CONF.register_opts(self.driver_opts,
                               constants.config_group(self.type))

        self.config = cfg.CONF.get(constants.config_group(self.type))

        assert issubclass(self.dataplane_instance_class, VPNInstanceDataplane)

        self.local_address = self.config.get("dataplane_local_address")

        if self.local_address is None:
            self.local_address = cfg.CONF.BGP.local_address

        self.log.info("Will use %s as local_address", self.local_address)

        # Linux kernel version check
        o = self._run_command("uname -r")
        self.kernel_release = o[0][0].split("-")[0]
        if self.required_kernel:
            if (version.StrictVersion(self.kernel_release) <
                    version.StrictVersion(self.required_kernel)):
                self.log.warning("%s requires at least Linux kernel %s"
                                 " (you are running %s)",
                                 self.__class__.__name__,
                                 self.required_kernel,
                                 self.kernel_release)

        # Flag to trigger cleanup all dataplane states on first call to
        # vif_plugged
        self.first_init = True

    @abc.abstractmethod
    def reset_state(self):
        '''dataplane cleanup hook (abstract)

        This method is called:
        - right before the work on the first attachment begins (initialize()
        is called afterwards)
        - by bagpipe-bgp-cleanup
        '''
        pass

    def initialize(self):
        '''dataplane initialization hook

        This is called after reset_state (which, e.g. cleans up the stuff
        possibly left-out by a previous run).

        Things that are reverted by reset_state() should go here.
        '''
        pass

    @log_decorator.log_info
    def initialize_dataplane_instance(self, instance_id, external_instance_id,
                                      gateway_ip, mask,
                                      instance_label, **kwargs):
        '''per-VPN dataplane instanciation

        returns a VPNInstanceDataplane subclass
        after calling reset_state on the dataplane driver, if this is the first
        call to initialize_dataplane_instance
        '''

        if self.first_init:
            self.log.info("First VPN instance init, reinitializing dataplane"
                          " state")
            try:
                self.reset_state()
            except Exception as e:
                self.log.error("Exception while resetting state: %s", e)

            try:
                self.initialize()
            except Exception as e:
                self.log.error("Exception while initializing dataplane"
                               " state: %s", e)
                raise

            self.first_init = False
        else:
            self.log.debug("(not reinitializing dataplane state)")

        return self.dataplane_instance_class(self, instance_id,
                                             external_instance_id,
                                             gateway_ip, mask,
                                             instance_label, **kwargs)

    def get_local_address(self):
        return self.local_address

    def supported_encaps(self):
        return self.__class__.encaps

    def needs_cleanup_assist(self):
        '''Control per-route cleanup events

        A dataplane driver not able to cleanup all the states for a given
        VPN instance can return True here to receive artifical dataplane
        removal calls, such as remove_dataplane_for_remote_endpoint, for
        each state previously setup
        '''
        return False

    def validate_directions(self, direction):
        # by default, assume a driver only supports plugging in both directions
        # if a driver does not support forwarding only the traffic to the port,
        # (and hence omit forwarding traffic from the port), it should raise an
        # exception if directions is TO_PORT
        # if a driver does not support forwarding only the traffic from the
        # port (and hence omit forwarding traffic to the port), it should raise
        # an exception if directions is FROM_PORT
        if (direction is not None) and (direction != constants.BOTH):
            self.log.warning("Unsupported direction: %s", direction)
            raise exc.APIException("Unsupported direction: %s" % direction)

    def _run_command(self, command, run_as_root=False, *args, **kwargs):
        return run_command.run_command(self.log, command, run_as_root,
                                       *args, **kwargs)

    def get_lg_map(self):
        encaps = []
        for encap in self.supported_encaps():
            encaps.append(repr(encap))
        return {
            "name": (lg.VALUE, self.__class__.__name__),
            "local_address": (lg.VALUE, self.local_address),
            "supported_encaps": (lg.VALUE, encaps),
            "config": (lg.VALUE, utils.osloconfig_json_serialize(self.config)),
            "kernel_release": (lg.VALUE, self.kernel_release)
        }


@six.add_metaclass(abc.ABCMeta)
class VPNInstanceDataplane(lg.LookingGlassLocalLogger):

    @log_decorator.log_info
    def __init__(self, dataplane_driver, instance_id, external_instance_id,
                 gateway_ip, network_plen, instance_label=None, **kwargs):
        lg.LookingGlassLocalLogger.__init__(self, repr(instance_id))
        self.driver = dataplane_driver
        self.config = dataplane_driver.config
        self.instance_id = instance_id
        self.external_instance_id = external_instance_id
        self.gateway_ip = gateway_ip
        self.network_plen = network_plen
        self.instance_label = instance_label

    @abc.abstractmethod
    def cleanup(self):
        pass

    @abc.abstractmethod
    def vif_plugged(self, mac_address, ip_address_prefix, localport, label,
                    direction):
        pass

    @abc.abstractmethod
    def vif_unplugged(self, mac_address, ip_address_prefix, localport, label,
                      direction, last_endpoint=True):
        pass

    def update_fallback(self, fallback):
        if fallback is not None:
            self.log.warning("fallback specified (%s) but not supported by "
                             "driver, ignoring", fallback)

    @abc.abstractmethod
    def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
                                            nlri, encaps,
                                            lb_consistent_hash_order=0):
        pass

    @abc.abstractmethod
    def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
                                             nlri, encaps,
                                             lb_consistent_hash_order=0):
        pass

    def needs_cleanup_assist(self):
        '''Control per-route cleanup events

        A dataplane driver not able to cleanup all the states for a given
        VPN instance can return True here to receive artifical dataplane
        removal calls, such as remove_dataplane_for_remote_endpoint, for
        each state previously setup
        '''
        return self.driver.needs_cleanup_assist()

    def _run_command(self, command, run_as_root=False, *args, **kwargs):
        return self.driver._run_command(command, run_as_root, *args, **kwargs)

    # Looking glass info ####

    def get_lg_local_info(self, path_prefix):
        driver = {"id": self.driver.type,
                  "href": lg.get_absolute_path(
                      "DATAPLANE_DRIVERS", path_prefix, [self.driver.type])}
        return {
            "driver": driver,
        }


class DummyVPNInstanceDataplane(VPNInstanceDataplane):

    @log_decorator.log
    def __init__(self, *args, **kwargs):
        VPNInstanceDataplane.__init__(self, *args)

    @log_decorator.log
    def vif_plugged(self, *args, **kwargs):
        pass

    @log_decorator.log
    def vif_unplugged(self, *args, **kwargs):
        pass

    @log_decorator.log
    def update_fallback(self, *args, **kwargs):
        pass

    @log_decorator.log
    def setup_dataplane_for_remote_endpoint(self, *args, **kwargs):
        pass

    @log_decorator.log
    def remove_dataplane_for_remote_endpoint(self, *args, **kwargs):
        pass

    @log_decorator.log
    def cleanup(self, *args, **kwargs):
        pass


class DummyDataplaneDriver(DataplaneDriver):

    dataplane_instance_class = DummyVPNInstanceDataplane

    def __init__(self, *args):
        DataplaneDriver.__init__(self, *args)
        self.log.warning("Dummy dataplane driver, won't do anything useful")

    @log_decorator.log_info
    def initialize(self, *args, **kwargs):
        pass

    @log_decorator.log_info
    def reset_state(self, *args, **kwargs):
        pass
